import io
import json
import os
import tempfile
import unittest

from nodus.lsp.server import LanguageServer


def _encode_message(payload: dict) -> bytes:
    body = json.dumps(payload).encode("utf-8")
    return f"Content-Length: {len(body)}\r\n\r\n".encode("ascii") + body


def _decode_messages(data: bytes) -> list[dict]:
    out: list[dict] = []
    stream = io.BytesIO(data)
    while True:
        header = stream.readline()
        if not header:
            break
        if not header.strip():
            continue
        length = int(header.decode("ascii").split(":", 1)[1].strip())
        blank = stream.readline()
        if blank not in {b"\r\n", b"\n"}:
            raise AssertionError("Invalid LSP framing")
        body = stream.read(length)
        out.append(json.loads(body.decode("utf-8")))
    return out


def _run_session(messages: list[dict]) -> list[dict]:
    input_bytes = b"".join(_encode_message(message) for message in messages)
    output = io.BytesIO()
    server = LanguageServer(io.BytesIO(input_bytes), output)
    server.run()
    return _decode_messages(output.getvalue())


class LspServerTests(unittest.TestCase):
    def test_initialization_handshake(self):
        responses = _run_session(
            [
                {"jsonrpc": "2.0", "id": 1, "method": "initialize", "params": {}},
                {"jsonrpc": "2.0", "id": 2, "method": "shutdown", "params": {}},
                {"jsonrpc": "2.0", "method": "exit", "params": {}},
            ]
        )
        self.assertEqual(responses[0]["id"], 1)
        capabilities = responses[0]["result"]["capabilities"]
        self.assertEqual(capabilities["textDocumentSync"], 1)
        self.assertTrue(capabilities["hoverProvider"])
        self.assertTrue(capabilities["definitionProvider"])
        self.assertEqual(responses[1]["id"], 2)
        self.assertIsNone(responses[1]["result"])

    def test_invalid_code_publishes_diagnostics(self):
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "bad.nd")
            uri = f"file:///{path.replace(os.sep, '/')}"
            responses = _run_session(
                [
                    {"jsonrpc": "2.0", "id": 1, "method": "initialize", "params": {}},
                    {
                        "jsonrpc": "2.0",
                        "method": "textDocument/didOpen",
                        "params": {
                            "textDocument": {
                                "uri": uri,
                                "languageId": "nodus",
                                "version": 1,
                                "text": "let x =\n",
                            }
                        },
                    },
                    {"jsonrpc": "2.0", "id": 2, "method": "shutdown", "params": {}},
                    {"jsonrpc": "2.0", "method": "exit", "params": {}},
                ]
            )
        diagnostics = [message for message in responses if message.get("method") == "textDocument/publishDiagnostics"][0]
        self.assertEqual(diagnostics["params"]["uri"], uri)
        published = diagnostics["params"]["diagnostics"]
        self.assertEqual(len(published), 1)
        self.assertIn("Unexpected token", published[0]["message"])
        self.assertEqual(published[0]["severity"], 1)
        self.assertIn("line", published[0])
        self.assertIn("column", published[0])

    def test_completion_returns_keywords_and_symbols(self):
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "main.nd")
            uri = f"file:///{path.replace(os.sep, '/')}"
            source = (
                'import "std:strings" as s\n'
                "fn greet(name: string) -> string {\n"
                "    return name\n"
                "}\n"
                'let local_name = greet("Nodus")\n'
                "print(lo)\n"
                "\n"
            )
            responses = _run_session(
                [
                    {"jsonrpc": "2.0", "id": 1, "method": "initialize", "params": {}},
                    {
                        "jsonrpc": "2.0",
                        "method": "textDocument/didOpen",
                        "params": {
                            "textDocument": {
                                "uri": uri,
                                "languageId": "nodus",
                                "version": 1,
                                "text": source,
                            }
                        },
                    },
                    {
                        "jsonrpc": "2.0",
                        "id": 2,
                        "method": "textDocument/completion",
                        "params": {
                            "textDocument": {"uri": uri},
                            "position": {"line": 5, "character": 8},
                        },
                    },
                    {
                        "jsonrpc": "2.0",
                        "id": 3,
                        "method": "textDocument/completion",
                        "params": {
                            "textDocument": {"uri": uri},
                            "position": {"line": 6, "character": 0},
                        },
                    },
                    {"jsonrpc": "2.0", "id": 4, "method": "shutdown", "params": {}},
                    {"jsonrpc": "2.0", "method": "exit", "params": {}},
                ]
            )
        symbol_completion = [message for message in responses if message.get("id") == 2][0]["result"]["items"]
        keyword_completion = [message for message in responses if message.get("id") == 3][0]["result"]["items"]
        symbol_labels = {item["label"] for item in symbol_completion}
        keyword_labels = {item["label"] for item in keyword_completion}
        self.assertIn("local_name", symbol_labels)
        self.assertIn("greet", keyword_labels)
        self.assertIn("s", keyword_labels)
        self.assertIn("let", keyword_labels)

    def test_definition_resolves_imported_symbol(self):
        with tempfile.TemporaryDirectory() as td:
            lib_path = os.path.join(td, "lib.nd")
            main_path = os.path.join(td, "main.nd")
            with open(lib_path, "w", encoding="utf-8") as handle:
                handle.write("export fn greet(name: string) -> string { return name }\n")
            uri = f"file:///{main_path.replace(os.sep, '/')}"
            source = 'import { greet } from "./lib.nd"\nprint(greet("Nodus"))\n'
            responses = _run_session(
                [
                    {"jsonrpc": "2.0", "id": 1, "method": "initialize", "params": {}},
                    {
                        "jsonrpc": "2.0",
                        "method": "textDocument/didOpen",
                        "params": {
                            "textDocument": {
                                "uri": uri,
                                "languageId": "nodus",
                                "version": 1,
                                "text": source,
                            }
                        },
                    },
                    {
                        "jsonrpc": "2.0",
                        "id": 2,
                        "method": "textDocument/definition",
                        "params": {
                            "textDocument": {"uri": uri},
                            "position": {"line": 1, "character": 7},
                        },
                    },
                    {"jsonrpc": "2.0", "id": 3, "method": "shutdown", "params": {}},
                    {"jsonrpc": "2.0", "method": "exit", "params": {}},
                ]
            )
        definition = [message for message in responses if message.get("id") == 2][0]["result"]
        self.assertTrue(definition["uri"].endswith("/lib.nd"))
        self.assertEqual(definition["range"]["start"]["line"], 0)


if __name__ == "__main__":
    unittest.main()



import unittest

from nodus.tooling.repl import (
    HELP_TEXT,
    ReplCommand,
    ReplState,
    describe_runtime_type,
    execute_repl_command,
    format_expression_ast,
    is_complete_chunk,
    parse_repl_command,
)


def make_state() -> ReplState:
    return ReplState(globals={}, fn_defs={}, import_state={})


class ReplCommandTests(unittest.TestCase):
    def test_parse_repl_command(self):
        command = parse_repl_command(":ast 1 + 2")
        self.assertEqual(command, ReplCommand(name="ast", arg="1 + 2"))
        self.assertIsNone(parse_repl_command("1 + 2"))

    def test_multiline_completion_tracks_braces(self):
        self.assertFalse(is_complete_chunk(["fn add(a, b) {", "return a + b"]))
        self.assertTrue(is_complete_chunk(["fn add(a, b) {", "return a + b", "}"]))

    def test_help_command(self):
        handled, output, should_exit = execute_repl_command(make_state(), ":help")
        self.assertTrue(handled)
        self.assertEqual(output, HELP_TEXT)
        self.assertFalse(should_exit)

    def test_ast_output(self):
        output = format_expression_ast("1 + 2 * 3")
        self.assertEqual(
            output,
            "\n".join(
                [
                    "Binary(+)",
                    "  Number(1)",
                    "  Binary(*)",
                    "    Number(2)",
                    "    Number(3)",
                ]
            ),
        )

    def test_ast_command(self):
        handled, output, should_exit = execute_repl_command(make_state(), ":ast 1 + 2")
        self.assertTrue(handled)
        self.assertIn("Binary(+)", output)
        self.assertFalse(should_exit)

    def test_dis_command(self):
        handled, output, should_exit = execute_repl_command(make_state(), ":dis 1 + 2")
        self.assertTrue(handled)
        self.assertIn("PUSH_CONST 1.0", output)
        self.assertIn("PUSH_CONST 2.0", output)
        self.assertIn("ADD", output)
        self.assertIn("RETURN", output)
        self.assertFalse(should_exit)

    def test_type_command_list(self):
        handled, output, should_exit = execute_repl_command(make_state(), ":type [1, 2, 3]")
        self.assertTrue(handled)
        self.assertEqual(output, "List<number>")
        self.assertFalse(should_exit)

    def test_type_command_map(self):
        handled, output, should_exit = execute_repl_command(make_state(), ':type ({"a": 1})')
        self.assertTrue(handled)
        self.assertEqual(output, "Map<string,number>")
        self.assertFalse(should_exit)

    def test_describe_runtime_type_mixed_list(self):
        self.assertEqual(describe_runtime_type([1, "x"]), "List<mixed>")


if __name__ == "__main__":
    unittest.main()



import os
import unittest

from nodus.tooling.formatter import format_source


FIXTURE_DIR = os.path.join(os.path.dirname(__file__), "fixtures", "fmt")


def load_fixture(name: str, kind: str) -> str:
    path = os.path.join(FIXTURE_DIR, f"{name}_{kind}.nd")
    with open(path, "r", encoding="utf-8") as f:
        return f.read()


class FormatterFixtureTests(unittest.TestCase):
    def test_formatter_fixtures(self):
        cases = [
            ("fmt_unary_literals", False),
            ("fmt_unary_grouping", False),
            ("fmt_comment_heavy", False),
            ("fmt_trailing_comments", False),
            ("fmt_trailing_comments_keep", True),
            ("fmt_import_export", False),
            ("fmt_nested_blocks", False),
            ("fmt_import_export_comments", False),
            ("fmt_import_export_comments_keep", True),
        ]
        for name, keep_trailing in cases:
            with self.subTest(name=name, keep_trailing=keep_trailing):
                src = load_fixture(name, "input")
                expected = load_fixture(name, "expected")
                formatted = format_source(src, keep_trailing_comments=keep_trailing)
                self.assertEqual(formatted, expected)
                self.assertEqual(format_source(formatted, keep_trailing_comments=keep_trailing), expected)

    def test_nested_unary_expressions(self):
        src = load_fixture("fmt_unary_grouping", "input")
        formatted = format_source(src)
        self.assertIn("let y = - -3", formatted)
        self.assertIn("let z = - -5", formatted)
        self.assertIn("print(- - -2)", formatted)

    def test_unary_expressions_with_comments(self):
        src = load_fixture("fmt_comment_heavy", "input")
        formatted = format_source(src)
        self.assertIn("let y = -(x + 1)", formatted)
        self.assertIn("# trail y", formatted)

    def test_trailing_comments_near_unary(self):
        src = load_fixture("fmt_trailing_comments", "input")
        formatted = format_source(src)
        self.assertIn("let x = -5", formatted)
        self.assertIn("// trailing unary", formatted)
        self.assertNotIn("let x = -5 // trailing unary", formatted)

        src_keep = load_fixture("fmt_trailing_comments_keep", "input")
        formatted_keep = format_source(src_keep, keep_trailing_comments=True)
        self.assertIn("let x = -5 // trailing unary", formatted_keep)
        self.assertIn("let y = -(x + 1) # grouped unary", formatted_keep)

    def test_numeric_literal_fidelity(self):
        src = load_fixture("fmt_unary_literals", "input")
        formatted = format_source(src)
        self.assertIn("let c = -2.50", formatted)
        self.assertIn("let list = [-1, -2.0, -3.75]", formatted)

    def test_idempotence(self):
        src = load_fixture("fmt_import_export", "input")
        formatted = format_source(src)
        self.assertEqual(format_source(formatted), formatted)


if __name__ == "__main__":
    unittest.main()



"""Math builtin functions for the Nodus VM."""

import math as _math
import random


def register(vm, registry) -> None:
    """Register numeric/math builtins onto the registry."""

    def builtin_math_abs(value):
        return abs(vm.ensure_number(value, "math_abs(x)"))

    def builtin_math_min(a, b):
        vm.ensure_number(a, "math_min(a, b)")
        vm.ensure_number(b, "math_min(a, b)")
        return min(a, b)

    def builtin_math_max(a, b):
        vm.ensure_number(a, "math_max(a, b)")
        vm.ensure_number(b, "math_max(a, b)")
        return max(a, b)

    def builtin_math_floor(value):
        return float(_math.floor(vm.ensure_number(value, "math_floor(x)")))

    def builtin_math_ceil(value):
        return float(_math.ceil(vm.ensure_number(value, "math_ceil(x)")))

    def builtin_math_sqrt(value):
        number = vm.ensure_number(value, "math_sqrt(x)")
        if number < 0:
            vm.runtime_error("runtime", "math_sqrt(x) expects a non-negative number")
        return _math.sqrt(number)

    def builtin_math_random():
        return random.random()

    registry.add("math_abs", 1, builtin_math_abs)
    registry.add("math_min", 2, builtin_math_min)
    registry.add("math_max", 2, builtin_math_max)
    registry.add("math_floor", 1, builtin_math_floor)
    registry.add("math_ceil", 1, builtin_math_ceil)
    registry.add("math_sqrt", 1, builtin_math_sqrt)
    registry.add("math_random", 0, builtin_math_random)



import http.client
import json
import time
import unittest

from nodus.services.server import run_in_thread


class ServerTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server, cls.thread = run_in_thread("127.0.0.1", 0, allowed_paths=["."])
        cls.port = cls.server.server_address[1]
        time.sleep(0.05)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server.server_close()

    def request(self, method: str, path: str, payload: dict | None = None):
        conn = http.client.HTTPConnection("127.0.0.1", self.port, timeout=5)
        body = None
        headers = {}
        if payload is not None:
            body = json.dumps(payload)
            headers["Content-Type"] = "application/json"
        conn.request(method, path, body=body, headers=headers)
        resp = conn.getresponse()
        data = resp.read().decode("utf-8")
        conn.close()
        return resp.status, json.loads(data) if data else {}

    def test_health(self):
        status, payload = self.request("GET", "/health")
        self.assertEqual(status, 200)
        self.assertEqual(payload["status"], "ok")
        self.assertEqual(payload["runtime"], "nodus")

    def test_execute(self):
        status, payload = self.request("POST", "/execute", {"code": "print(1 + 1)", "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertEqual(payload["stdout"], "2.0\n")

    def test_syntax_error(self):
        status, payload = self.request("POST", "/execute", {"code": "fn {", "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertFalse(payload["ok"])
        self.assertEqual(payload["error"]["type"], "syntax")

    def test_ast(self):
        status, payload = self.request("POST", "/ast", {"code": "let x = 1", "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertIn("Module", payload["ast_pretty"])
        self.assertEqual(payload["ast"][0]["type"], "Let")

    def test_sandbox_timeout(self):
        status, payload = self.request("POST", "/execute", {"code": "while (true) { }", "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertFalse(payload["ok"])
        self.assertEqual(payload["error"]["type"], "sandbox")
        self.assertIn("timed out", payload["error"]["message"])

    def test_input_blocked(self):
        status, payload = self.request("POST", "/execute", {"code": "input(\"x\")", "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertFalse(payload["ok"])
        self.assertEqual(payload["error"]["type"], "sandbox")
        self.assertIn("input()", payload["error"]["message"])


class ServerAuthTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.token = "secret-token"
        cls.server, cls.thread = run_in_thread("127.0.0.1", 0, allowed_paths=["."], auth_token=cls.token)
        cls.port = cls.server.server_address[1]
        time.sleep(0.05)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server.server_close()

    def request(self, method: str, path: str, payload: dict | None = None, *, auth: bool = False):
        conn = http.client.HTTPConnection("127.0.0.1", self.port, timeout=5)
        body = None
        headers = {}
        if payload is not None:
            body = json.dumps(payload)
            headers["Content-Type"] = "application/json"
        if auth:
            headers["Authorization"] = f"Bearer {self.token}"
        conn.request(method, path, body=body, headers=headers)
        resp = conn.getresponse()
        data = resp.read().decode("utf-8")
        conn.close()
        return resp.status, json.loads(data) if data else {}

    def test_auth_required(self):
        status, payload = self.request("GET", "/health")
        self.assertEqual(status, 401)
        self.assertEqual(payload.get("error"), "unauthorized")

    def test_auth_allows_access(self):
        status, payload = self.request("GET", "/health", auth=True)
        self.assertEqual(status, 200)
        self.assertEqual(payload["status"], "ok")


if __name__ == "__main__":
    unittest.main()



"""Tool wrapper for AST generation."""

from __future__ import annotations

from nodus.result import Result
from nodus.tooling.runner import build_ast


TOOL_SPEC = {
    "name": "nodus_ast",
    "description": "Build a Nodus AST and return both pretty and structured forms.",
    "parameters": {
        "type": "object",
        "properties": {
            "code": {"type": "string"},
            "filename": {"type": "string"},
        },
        "required": ["code"],
    },
}


def _input_error(message: str, filename: str | None) -> dict:
    return Result.failure(
        stage="ast",
        filename=filename,
        stdout="",
        stderr="",
        errors=[{"type": "InputError", "message": message}],
        error={"type": "input", "message": message, "path": filename},
    ).to_dict()


def nodus_ast(code: str, filename: str = "<memory>") -> dict:
    if not isinstance(code, str):
        return _input_error("code must be a string", filename)
    if filename is not None and not isinstance(filename, str):
        return _input_error("filename must be a string", filename)
    return build_ast(code, filename=filename)



import io
import json
import os
import tempfile
import time
import unittest
from contextlib import redirect_stdout

import nodus as lang
from nodus.runtime.profiler import Profiler


class ProfilerTests(unittest.TestCase):
    def test_profiler_records_opcodes(self):
        profiler = Profiler()
        profiler.start()
        profiler.record_opcode("LOAD_CONST")
        profiler.record_opcode("LOAD_CONST")
        profiler.record_opcode("ADD")
        profiler.stop()
        self.assertEqual(profiler.opcode_counts.get("LOAD_CONST"), 2)
        self.assertEqual(profiler.opcode_counts.get("ADD"), 1)

    def test_profiler_counts_function_calls(self):
        profiler = Profiler()
        profiler.start()
        profiler.record_function_call("main")
        profiler.record_function_call("main")
        profiler.record_function_call("process")
        profiler.stop()
        self.assertEqual(profiler.function_calls.get("main"), 2)
        self.assertEqual(profiler.function_calls.get("process"), 1)

    def test_profiler_records_time(self):
        profiler = Profiler()
        profiler.start()
        profiler.enter_function("main")
        time.sleep(0.001)
        profiler.exit_function("main")
        profiler.stop()
        report = profiler.report()
        self.assertGreaterEqual(report.get("total_time_ms", 0.0), 0.0)
        functions = report.get("functions", [])
        main_entry = next((item for item in functions if item.get("name") == "main"), None)
        self.assertIsNotNone(main_entry)
        self.assertGreaterEqual(main_entry.get("time_ms", 0.0), 0.0)

    def test_cli_profile_command(self):
        code = """
fn add(a, b) {
    return a + b
}

add(1, 2)
"""
        with tempfile.TemporaryDirectory() as td:
            script = os.path.join(td, "profile.nd")
            with open(script, "w", encoding="utf-8") as f:
                f.write(code)

            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "profile", script])
            self.assertEqual(exit_code, 0)
            output = buf.getvalue()
            self.assertIn("Nodus Profiling Report", output)

            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "profile", script, "--json"])
            self.assertEqual(exit_code, 0)
            payload = json.loads(buf.getvalue().strip())
            self.assertIn("runtime_ms", payload)
            self.assertIn("functions", payload)
            self.assertIn("opcodes", payload)


if __name__ == "__main__":
    unittest.main()



"""REPL support for Nodus."""

from __future__ import annotations

from dataclasses import dataclass
import os

from nodus.builtins.nodus_builtins import BUILTIN_NAMES
from nodus.compiler.compiler import Compiler, format_bytecode, wrap_bytecode
from nodus.frontend.ast.ast_nodes import (
    Attr,
    Bin,
    Bool,
    Call,
    ExprStmt,
    FnDef,
    Index,
    ListLit,
    MapLit,
    Nil,
    Num,
    Str,
    Unary,
    Var,
)
from nodus.frontend.lexer import tokenize
from nodus.frontend.parser import Parser
from nodus.orchestration.task_graph import TaskGraph, TaskNode
from nodus.runtime.coroutine import Coroutine
from nodus.runtime.channel import Channel
from nodus.runtime.diagnostics import format_error
from nodus.tooling.loader import set_module_on_tree
from nodus.runtime.module_loader import ModuleLoader
from nodus.vm.vm import Closure, Record, VM

try:
    import readline
except ImportError:  # pragma: no cover - platform dependent
    readline = None


HISTORY_FILE = os.path.expanduser("~/.nodus_history")
HELP_TEXT = "\n".join(
    [
        "Nodus REPL Commands",
        "",
        ":ast <expr>    show AST",
        ":dis <expr>    show bytecode",
        ":type <expr>   show inferred type",
        ":help          show commands",
        ":quit          exit REPL",
    ]
)
TYPE_TEMP_NAME = "__repl_type_value__"


@dataclass
class ReplState:
    globals: dict
    fn_defs: dict[str, FnDef]
    import_state: dict


@dataclass
class ReplCommand:
    name: str
    arg: str


def split_fn_defs(stmts: list):
    fn_defs = {}
    non_fn = []
    for stmt in stmts:
        if isinstance(stmt, FnDef):
            fn_defs[stmt.name] = stmt
        else:
            non_fn.append(stmt)
    return fn_defs, non_fn


def is_complete_chunk(lines: list[str]) -> bool:
    depth = 0
    for line in lines:
        depth += line.count("{")
        depth -= line.count("}")
    return depth <= 0


def parse_repl_command(src: str) -> ReplCommand | None:
    stripped = src.strip()
    if not stripped.startswith(":"):
        return None
    body = stripped[1:].strip()
    if not body:
        raise ValueError("Empty REPL command. Use :help for available commands.")
    name, _, arg = body.partition(" ")
    return ReplCommand(name=name.lower(), arg=arg.strip())


def _setup_history() -> None:
    if readline is None:
        return
    try:
        readline.read_history_file(HISTORY_FILE)
    except FileNotFoundError:
        pass
    readline.set_history_length(1000)


def _save_history() -> None:
    if readline is None:
        return
    try:
        readline.write_history_file(HISTORY_FILE)
    except OSError:
        pass


def _parse_repl_ast(src: str):
    stmts = Parser(tokenize(src)).parse()
    set_module_on_tree(stmts, "<repl>")
    return stmts


def _parse_repl_expression(src: str):
    stmts = _parse_repl_ast(src)
    if len(stmts) != 1 or not isinstance(stmts[0], ExprStmt):
        raise ValueError("Expected a single expression.")
    return stmts[0].expr


def format_expression_ast(src: str) -> str:
    expr = _parse_repl_expression(src)
    lines: list[str] = []
    _format_expr_node(expr, 0, lines)
    return "\n".join(lines)


def _format_expr_node(node, indent: int, lines: list[str]) -> None:
    prefix = "  " * indent
    if isinstance(node, Num):
        value = node.raw if node.raw is not None else str(node.v)
        lines.append(f"{prefix}Number({value})")
        return
    if isinstance(node, Bool):
        lines.append(f"{prefix}Bool({str(node.v).lower()})")
        return
    if isinstance(node, Str):
        lines.append(f'{prefix}String("{node.v}")')
        return
    if isinstance(node, Nil):
        lines.append(f"{prefix}Nil")
        return
    if isinstance(node, Var):
        lines.append(f"{prefix}Identifier({node.name})")
        return
    if isinstance(node, Unary):
        lines.append(f"{prefix}Unary({node.op})")
        _format_expr_node(node.expr, indent + 1, lines)
        return
    if isinstance(node, Bin):
        lines.append(f"{prefix}Binary({node.op})")
        _format_expr_node(node.a, indent + 1, lines)
        _format_expr_node(node.b, indent + 1, lines)
        return
    if isinstance(node, ListLit):
        lines.append(f"{prefix}List")
        for item in node.items:
            _format_expr_node(item, indent + 1, lines)
        return
    if isinstance(node, MapLit):
        lines.append(f"{prefix}Map")
        for key, value in node.items:
            lines.append(f"{prefix}  Entry")
            _format_expr_node(key, indent + 2, lines)
            _format_expr_node(value, indent + 2, lines)
        return
    if isinstance(node, Attr):
        lines.append(f"{prefix}Attr({node.name})")
        _format_expr_node(node.obj, indent + 1, lines)
        return
    if isinstance(node, Index):
        lines.append(f"{prefix}Index")
        _format_expr_node(node.seq, indent + 1, lines)
        _format_expr_node(node.index, indent + 1, lines)
        return
    if isinstance(node, Call):
        lines.append(f"{prefix}Call")
        _format_expr_node(node.callee, indent + 1, lines)
        for arg in node.args:
            _format_expr_node(arg, indent + 1, lines)
        return
    lines.append(f"{prefix}{type(node).__name__}")


def _build_compiler() -> Compiler:
    return Compiler(module_infos={"<repl>": None}, module_defs_index={}, builtin_names=BUILTIN_NAMES)


def _compile_repl_program(state: ReplState, stmts: list):
    program = list(state.fn_defs.values()) + stmts
    compiler = Compiler(module_infos={"<repl>": None}, module_defs_index={}, builtin_names=BUILTIN_NAMES)
    code, functions, code_locs = compiler.compile_program(program)
    return wrap_bytecode(code, module_name="<repl>"), functions, code_locs


def _compile_repl_expression(state: ReplState, src: str):
    expr = _parse_repl_expression(src)
    compiler = _build_compiler()
    jump_to_main = compiler.emit("JUMP", None)
    compiler.current_module = "<repl>"
    fn_defs = list(state.fn_defs.values())
    compiler.predeclare_module_scopes(fn_defs)
    for fn in fn_defs:
        compiler.compile_fn_def(fn)
    main_addr = len(compiler.code)
    compiler.patch(jump_to_main, "JUMP", main_addr)
    compiler.compile_expr(expr)
    compiler.emit("RETURN")
    return compiler.code, compiler.functions, compiler.code_locs, main_addr


def disassemble_expression(state: ReplState, src: str) -> str:
    code, functions, code_locs, main_addr = _compile_repl_expression(state, src)
    text = format_bytecode(code, code_locs, functions)
    lines = []
    in_main = False
    for line in text.splitlines():
        if line == "Function main:":
            in_main = True
            continue
        if in_main and line.startswith("Function "):
            break
        if not in_main:
            continue
        body = line.strip()
        if ": " in body:
            body = body.split(": ", 1)[1]
        if "  (" in body and body.endswith(")"):
            body = body.rsplit("  (", 1)[0]
        if body.endswith("HALT"):
            continue
        lines.append(body)
    if not lines:
        for instr in code[main_addr:]:
            op = instr[0]
            if op == "HALT":
                continue
            operands = " ".join(repr(value) if not isinstance(value, str) else value for value in instr[1:])
            lines.append(f"{op} {operands}".strip())
    return "\n".join(lines)


def infer_expression_type(state: ReplState, src: str) -> str:
    expr = _parse_repl_expression(src)
    temp_stmt = Parser(tokenize(f"let {TYPE_TEMP_NAME} = 0")).parse()[0]
    temp_stmt.expr = expr
    set_module_on_tree([temp_stmt], "<repl>")
    bytecode, functions, code_locs = _compile_repl_program(state, [temp_stmt])
    vm = VM(
        bytecode,
        functions,
        code_locs=code_locs,
        module_globals=state.globals,
        source_path="<repl>",
    )
    vm.run()
    try:
        return describe_runtime_type(vm.module_globals.get(TYPE_TEMP_NAME))
    finally:
        vm.module_globals.pop(TYPE_TEMP_NAME, None)


def describe_runtime_type(value) -> str:
    if value is None:
        return "nil"
    if isinstance(value, bool):
        return "bool"
    if isinstance(value, (int, float)) and not isinstance(value, bool):
        return "number"
    if isinstance(value, str):
        return "string"
    if isinstance(value, list):
        inner = _merge_types(describe_runtime_type(item) for item in value)
        return f"List<{inner}>"
    if isinstance(value, dict):
        key_type = _merge_types(describe_runtime_type(item) for item in value.keys())
        value_type = _merge_types(describe_runtime_type(item) for item in value.values())
        return f"Map<{key_type},{value_type}>"
    if isinstance(value, Record):
        return value.kind
    if isinstance(value, Closure):
        return "function"
    if isinstance(value, Coroutine):
        return "coroutine"
    if isinstance(value, Channel):
        return "channel"
    if isinstance(value, TaskNode):
        return "task"
    if isinstance(value, TaskGraph):
        return "graph"
    return "unknown"


def _merge_types(values) -> str:
    types = sorted(set(values))
    if not types:
        return "any"
    if len(types) == 1:
        return types[0]
    return "mixed"


def execute_repl_command(state: ReplState, src: str) -> tuple[bool, str | None, bool]:
    command = parse_repl_command(src)
    if command is None:
        return False, None, False
    if command.name == "help":
        return True, HELP_TEXT, False
    if command.name == "quit":
        return True, None, True
    if command.name == "ast":
        if not command.arg:
            raise ValueError("Usage: :ast <expr>")
        return True, format_expression_ast(command.arg), False
    if command.name == "dis":
        if not command.arg:
            raise ValueError("Usage: :dis <expr>")
        return True, disassemble_expression(state, command.arg), False
    if command.name == "type":
        if not command.arg:
            raise ValueError("Usage: :type <expr>")
        return True, infer_expression_type(state, command.arg), False
    raise ValueError(f"Unknown REPL command: :{command.name}")


def _execute_source(state: ReplState, loader: ModuleLoader, src: str) -> None:
    stmts = _parse_repl_ast(src)
    metadata = loader._build_metadata("<repl>", base_dir=os.getcwd(), source=src, source_path=None)
    bindings, _deps = loader._resolve_import_bindings(metadata)
    state.globals.update(bindings)
    new_defs, non_fn = split_fn_defs(stmts)
    merged_defs = dict(state.fn_defs)
    merged_defs.update(new_defs)

    bytecode, functions, code_locs = _compile_repl_program(
        ReplState(globals=state.globals, fn_defs=merged_defs, import_state=state.import_state),
        non_fn,
    )
    vm = VM(
        bytecode,
        functions,
        code_locs=code_locs,
        module_globals=state.globals,
        source_path="<repl>",
    )
    vm.run()
    state.globals = vm.module_globals
    state.fn_defs = merged_defs


def run_repl(version: str):
    state = ReplState(
        globals={},
        fn_defs={},
        import_state={"loaded": set(), "loading": set(), "exports": {}, "modules": {}, "module_ids": {}, "project_root": None},
    )
    loader = ModuleLoader(project_root=os.getcwd())
    _setup_history()
    print(f"{version} REPL (type 'exit', 'quit', or ':quit' to quit)")

    try:
        while True:
            lines = []
            prompt = "> "

            while True:
                try:
                    line = input(prompt)
                except EOFError:
                    print()
                    return

                if not lines and line.strip() in {"exit", "quit"}:
                    return

                lines.append(line)
                if is_complete_chunk(lines):
                    break
                prompt = "... "

            src = "\n".join(lines).strip()
            if not src:
                continue

            try:
                handled, output, should_exit = execute_repl_command(state, src)
                if handled:
                    if output:
                        print(output)
                    if should_exit:
                        return
                    continue
                _execute_source(state, loader, src)
            except Exception as err:
                print(format_error(err))
    finally:
        _save_history()



import unittest

from nodus.runtime.embedding import NodusRuntime
from nodus.runtime.errors import NodusRuntimeError, NodusSandboxError


class EmbeddingInputTests(unittest.TestCase):
    def test_input_blocked_by_default(self):
        runtime = NodusRuntime()
        with self.assertRaises(NodusSandboxError) as ctx:
            runtime.run_source('input("x")', filename="inline.nd")
        self.assertEqual(ctx.exception.error_type, "SandboxError")
        self.assertIn("input()", str(ctx.exception))

    def test_input_allowed_when_enabled(self):
        runtime = NodusRuntime(allow_input=True)
        runtime.register_function("host_input", lambda: "ok", arity=0)
        result = runtime.run_source('print(host_input())', filename="inline.nd")
        self.assertTrue(result["ok"])
        self.assertEqual(result["stdout"], "ok\n")


if __name__ == "__main__":
    unittest.main()



"""FastAPI execution service for Nodus runner functions."""

from __future__ import annotations

from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse

from nodus.tooling.runner import (
    agent_call_result,
    build_ast,
    check_source,
    disassemble_source,
    memory_delete_result,
    memory_get_result,
    memory_keys_result,
    memory_put_result,
    plan_graph_source,
    resume_graph,
    run_source,
    tool_call_result,
)
from nodus.orchestration.task_graph import latest_graph_state, load_graph_state


class ExecutionState:
    def __init__(self):
        self.last_vm = None

    def _graph_metadata(self, vm, graph_id: str | None = None) -> dict:
        resolved_id = graph_id
        if resolved_id is None and vm is not None and getattr(vm, "last_graph_plan", None):
            resolved_id = vm.last_graph_plan.get("graph_id")

        if resolved_id is None:
            resolved_id, state = latest_graph_state()
        else:
            state = load_graph_state(resolved_id)

        tasks = state.get("tasks", {}) if state else {}
        status = state.get("status") if state else None
        return {"graph_id": resolved_id, "tasks": tasks, "graph_status": status}

    def execute(self, payload: dict) -> dict:
        code = payload.get("code", "")
        filename = payload.get("filename")
        result, vm = run_source(code, filename=filename)
        self.last_vm = vm
        return result

    def check(self, payload: dict) -> dict:
        code = payload.get("code", "")
        filename = payload.get("filename")
        return check_source(code, filename=filename)

    def ast(self, payload: dict) -> dict:
        code = payload.get("code", "")
        filename = payload.get("filename")
        compact = bool(payload.get("compact", False))
        return build_ast(code, filename=filename, compact=compact)

    def disassemble(self, payload: dict) -> dict:
        code = payload.get("code", "")
        filename = payload.get("filename")
        return disassemble_source(code, filename=filename)

    def plan_graph(self, payload: dict) -> dict:
        code = payload.get("code", "")
        filename = payload.get("filename")
        result, vm = plan_graph_source(code, filename=filename)
        self.last_vm = vm
        if result.get("ok"):
            result.update(self._graph_metadata(vm, result.get("plan", {}).get("graph_id")))
        return result

    def resume_graph(self, payload: dict) -> dict:
        graph_id = payload.get("graph_id")
        if self.last_vm is not None:
            from nodus.tooling.runner import resume_graph_in_vm
            result, vm = resume_graph_in_vm(self.last_vm, graph_id)
        else:
            result, vm = resume_graph(graph_id)
        self.last_vm = vm
        if result.get("ok"):
            result.update(self._graph_metadata(vm, graph_id))
        return result

    def graph_run(self, payload: dict) -> dict:
        code = payload.get("code", "")
        filename = payload.get("filename")
        result, vm = run_source(code, filename=filename)
        self.last_vm = vm
        result.update(self._graph_metadata(vm))
        return result

    def graph_plan(self, payload: dict) -> dict:
        return self.plan_graph(payload)

    def graph_resume(self, payload: dict) -> dict:
        return self.resume_graph(payload)

    def runtime_events(self) -> dict:
        vm = self.last_vm
        events = [event.to_dict() for event in vm.event_bus.events()] if vm is not None else []
        return {"ok": True, "events": events}

    def tool_call(self, payload: dict) -> dict:
        return tool_call_result(payload.get("name"), payload.get("args", {}), vm=self.last_vm)

    def agent_call(self, payload: dict) -> dict:
        return agent_call_result(payload.get("name"), payload.get("payload"), vm=self.last_vm)

    def memory_get(self, key: str | None = None) -> dict:
        if key is None:
            return memory_keys_result(vm=self.last_vm)
        return memory_get_result(key, vm=self.last_vm)

    def memory_put(self, payload: dict) -> dict:
        return memory_put_result(payload.get("key"), payload.get("value"), vm=self.last_vm)

    def memory_delete(self, key: str) -> dict:
        return memory_delete_result(key, vm=self.last_vm)


def create_app(state: ExecutionState | None = None) -> FastAPI:
    state = state or ExecutionState()
    app = FastAPI()

    @app.post("/execute")
    async def execute(request: Request):
        payload = await request.json()
        return JSONResponse(state.execute(payload))

    @app.post("/check")
    async def check(request: Request):
        payload = await request.json()
        return JSONResponse(state.check(payload))

    @app.post("/ast")
    async def ast(request: Request):
        payload = await request.json()
        return JSONResponse(state.ast(payload))

    @app.post("/disassemble")
    async def disassemble(request: Request):
        payload = await request.json()
        return JSONResponse(state.disassemble(payload))

    @app.post("/dis")
    async def dis(request: Request):
        payload = await request.json()
        return JSONResponse(state.disassemble(payload))

    @app.post("/plan_graph")
    async def plan_graph(request: Request):
        payload = await request.json()
        return JSONResponse(state.plan_graph(payload))

    @app.post("/resume_graph")
    async def resume_graph_endpoint(request: Request):
        payload = await request.json()
        return JSONResponse(state.resume_graph(payload))

    @app.post("/graph/run")
    async def graph_run(request: Request):
        payload = await request.json()
        return JSONResponse(state.graph_run(payload))

    @app.post("/graph/plan")
    async def graph_plan(request: Request):
        payload = await request.json()
        return JSONResponse(state.graph_plan(payload))

    @app.post("/graph/resume")
    async def graph_resume(request: Request):
        payload = await request.json()
        return JSONResponse(state.graph_resume(payload))

    @app.get("/runtime/events")
    async def runtime_events():
        return JSONResponse(state.runtime_events())

    @app.post("/tool/call")
    async def tool_call(request: Request):
        payload = await request.json()
        return JSONResponse(state.tool_call(payload))

    @app.post("/agent/call")
    async def agent_call(request: Request):
        payload = await request.json()
        return JSONResponse(state.agent_call(payload))

    @app.get("/memory")
    async def memory(key: str | None = None):
        return JSONResponse(state.memory_get(key))

    @app.post("/memory")
    async def memory_put(request: Request):
        payload = await request.json()
        return JSONResponse(state.memory_put(payload))

    @app.delete("/memory/{key}")
    async def memory_delete(key: str):
        return JSONResponse(state.memory_delete(key))

    return app



from nodus.tooling.installer import install_project
from nodus.tooling.project import load_project, read_lockfile
from nodus.tooling.resolver import resolve_project_dependencies


def test_installer_copies_path_dependency_into_nodus_modules(tmp_path):
    dependency = tmp_path / "utils"
    dependency.mkdir()
    (dependency / "nodus.toml").write_text('name = "utils"\nversion = "1.2.3"\n', encoding="utf-8")
    (dependency / "strings.nd").write_text('export fn upper(value) { return value }\n', encoding="utf-8")

    project = tmp_path / "app"
    project.mkdir()
    (project / "nodus.toml").write_text(
        '\n'.join(
            [
                'name = "app"',
                'version = "0.1.0"',
                "",
                "[dependencies]",
                'utils = { path = "../utils" }',
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    config = load_project(str(project))
    resolution = resolve_project_dependencies(config, update=True)
    installed = install_project(config, resolution)

    assert (project / ".nodus" / "modules" / "utils" / "strings.nd").is_file()
    assert installed["utils"].version == "1.2.3"
    assert installed["utils"].hash.startswith("sha256:")
    assert read_lockfile(config.lock_path)["utils"].version == "1.2.3"



"""Tokenization and keyword handling for Nodus."""

import re
from dataclasses import dataclass

from nodus.runtime.diagnostics import LangSyntaxError


@dataclass
class Tok:
    kind: str
    val: str
    line: int
    col: int


TOKEN_RE = re.compile(
    r"""
    (?P<WS>[ \t]+)
  | (?P<COMMENT1>\#.*)
  | (?P<COMMENT2>//.*)
  | (?P<NL>\n+)
  | (?P<STR>"(?:\\.|[^"\\])*")
  | (?P<NUM>\d+(\.\d+)?)
  | (?P<ID>[A-Za-z_][A-Za-z0-9_]*)
  | (?P<OP>&&|\|\||==|!=|<=|>=|->|[+\-*/=(){}\[\],;:.<>!])
    """,
    re.VERBOSE,
)

KEYWORDS = {
    "let",
    "print",
    "if",
    "else",
    "while",
    "for",
    "true",
    "false",
    "fn",
    "return",
    "nil",
    "import",
    "as",
    "export",
    "from",
    "try",
    "catch",
    "finally",
    "throw",
    "record",
    "in",
    "yield",
    "workflow",
    "goal",
    "step",
    "after",
    "with",
    "action",
}

# Simple single-character escape sequences.
# \x and \u are handled separately in decode_string_literal because they
# consume additional hex-digit characters from the source text.
ESCAPE_MAP = {
    "n": "\n",
    "t": "\t",
    "r": "\r",
    "0": "\0",
    '"': '"',
    "\\": "\\",
}


def decode_string_literal(
    token_text: str,
    *,
    line: int | None = None,
    col: int | None = None,
) -> str:
    """Decode a quoted string token into its runtime string value.

    Raises LangSyntaxError (with source location when provided) for any
    malformed escape sequence so that callers do not need a try/except.

    Supported escape sequences:
        \\\\  backslash
        \\"   double quote
        \\n   newline (U+000A)
        \\t   horizontal tab (U+0009)
        \\r   carriage return (U+000D)
        \\0   null byte (U+0000)
        \\xHH two-digit hex byte  (e.g. \\x41 -> 'A')
        \\uXXXX four-digit Unicode code point (e.g. \\u03B1 -> 'α')
    """
    s = token_text[1:-1]
    out = []
    i = 0

    while i < len(s):
        ch = s[i]
        if ch != "\\":
            out.append(ch)
            i += 1
            continue

        i += 1
        if i >= len(s):
            raise LangSyntaxError(
                "Unterminated escape sequence in string literal",
                line=line,
                col=col,
            )

        esc = s[i]

        # \xHH — two-digit hex byte
        if esc == "x":
            hex_digits = s[i + 1 : i + 3]
            if len(hex_digits) < 2:
                raise LangSyntaxError(
                    r"Incomplete \x escape: expected 2 hex digits",
                    line=line,
                    col=col,
                )
            try:
                out.append(chr(int(hex_digits, 16)))
            except ValueError:
                raise LangSyntaxError(
                    rf"Invalid \x escape: \x{hex_digits}",
                    line=line,
                    col=col,
                )
            i += 3
            continue

        # \uXXXX — four-digit Unicode code point
        if esc == "u":
            hex_digits = s[i + 1 : i + 5]
            if len(hex_digits) < 4:
                raise LangSyntaxError(
                    r"Incomplete \u escape: expected 4 hex digits",
                    line=line,
                    col=col,
                )
            try:
                out.append(chr(int(hex_digits, 16)))
            except ValueError:
                raise LangSyntaxError(
                    rf"Invalid \u escape: \u{hex_digits}",
                    line=line,
                    col=col,
                )
            i += 5
            continue

        if esc not in ESCAPE_MAP:
            raise LangSyntaxError(
                f"Unsupported escape sequence: \\{esc}",
                line=line,
                col=col,
            )

        out.append(ESCAPE_MAP[esc])
        i += 1

    return "".join(out)


def tokenize(src: str) -> list[Tok]:
    if src.startswith("\ufeff"):
        src = src[1:]
    src = src.replace("\r\n", "\n").replace("\r", "\n")

    pos = 0
    line = 1
    col = 1
    out: list[Tok] = []

    while pos < len(src):
        start_line = line
        start_col = col
        m = TOKEN_RE.match(src, pos)
        if not m:
            raise LangSyntaxError(f"Unexpected character {src[pos]!r}", line=start_line, col=start_col)

        kind = m.lastgroup
        text = m.group(kind)
        pos = m.end()

        if kind in {"WS"}:
            col += len(text)
            continue
        if kind in {"COMMENT1", "COMMENT2"}:
            out.append(Tok("COMMENT", text, start_line, start_col))
            col += len(text)
            continue
        if kind == "NL":
            out.append(Tok("SEP", "\n", start_line, start_col))
            line += len(text)
            col = 1
            continue
        if kind == "NUM":
            out.append(Tok("NUM", text, start_line, start_col))
            col += len(text)
            continue
        if kind == "STR":
            # decode_string_literal raises LangSyntaxError directly; no try/except needed.
            val = decode_string_literal(text, line=start_line, col=start_col)
            out.append(Tok("STR", val, start_line, start_col))
            # A string token may span multiple lines if it contains a literal
            # newline character (matched by [^"\\] in the regex).  Update line
            # and col correctly so subsequent tokens have accurate positions.
            newlines_in_str = text.count("\n")
            if newlines_in_str > 0:
                line += newlines_in_str
                col = len(text) - text.rfind("\n")
            else:
                col += len(text)
            continue
        if kind == "ID":
            if text in KEYWORDS:
                out.append(Tok(text.upper(), text, start_line, start_col))
            else:
                out.append(Tok("ID", text, start_line, start_col))
            col += len(text)
            continue
        if kind == "OP":
            if text == ";":
                out.append(Tok("SEP", text, start_line, start_col))
            else:
                out.append(Tok(text, text, start_line, start_col))
            col += len(text)

    out.append(Tok("EOF", "", line, col))
    return out



"""Static type helpers for Nodus."""


class NodusType:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, other) -> bool:
        return isinstance(other, NodusType) and self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)

    def __repr__(self) -> str:
        return self.name


class FunctionType(NodusType):
    def __init__(self, params: list[NodusType], return_type: NodusType):
        super().__init__("function")
        self.params = params
        self.return_type = return_type

    def __repr__(self) -> str:
        params = ", ".join(param.name for param in self.params)
        return f"function({params}) -> {self.return_type.name}"


ANY = NodusType("any")
INT = NodusType("int")
FLOAT = NodusType("float")
STRING = NodusType("string")
BOOL = NodusType("bool")
LIST = NodusType("list")
RECORD = NodusType("record")
FUNCTION = NodusType("function")
NIL = NodusType("nil")

TYPE_NAMES = {
    "any": ANY,
    "int": INT,
    "float": FLOAT,
    "string": STRING,
    "bool": BOOL,
    "list": LIST,
    "record": RECORD,
    "function": FUNCTION,
}


def parse_type_name(name: str | None) -> NodusType:
    if name is None:
        return ANY
    return TYPE_NAMES.get(name, ANY)


def is_assignable(expected: NodusType, actual: NodusType) -> bool:
    if expected == ANY or actual == ANY:
        return True
    if expected == FLOAT and actual == INT:
        return True
    if expected == FUNCTION and isinstance(actual, FunctionType):
        return True
    return expected == actual


def combine_types(left: NodusType, right: NodusType) -> NodusType:
    if left == right:
        return left
    if left == ANY or right == ANY:
        return ANY
    if (left == INT and right == FLOAT) or (left == FLOAT and right == INT):
        return FLOAT
    return ANY



"""Collection, string, and JSON builtin functions for the Nodus VM."""

import json

from nodus.vm.runtime_values import is_json_safe


def register(vm, registry) -> None:
    """Register collection, string, and JSON builtins onto the registry."""

    def builtin_len(value):
        if isinstance(value, (str, list, dict)):
            return float(len(value))
        vm.runtime_error("type", "len(x) expects string, list, or map")

    def builtin_upper(value):
        vm.ensure_string(value, "upper(x)")
        return value.upper()

    def builtin_lower(value):
        vm.ensure_string(value, "lower(x)")
        return value.lower()

    def builtin_trim(value):
        vm.ensure_string(value, "trim(x)")
        return value.strip()

    def builtin_split(value, delimiter):
        vm.ensure_string(value, "split(x, delimiter)")
        vm.ensure_string(delimiter, "split(x, delimiter)")
        return value.split(delimiter)

    def builtin_contains(value, needle):
        vm.ensure_string(value, "contains(x, needle)")
        vm.ensure_string(needle, "contains(x, needle)")
        return needle in value

    def builtin_keys(value):
        if not isinstance(value, dict):
            vm.runtime_error("type", "keys(x) expects a map")
        return list(value.keys())

    def builtin_values(value):
        if not isinstance(value, dict):
            vm.runtime_error("type", "values(x) expects a map")
        return list(value.values())

    def builtin_list_push(value, item):
        if not isinstance(value, list):
            vm.runtime_error("type", "list_push(list, value) expects a list")
        value.append(item)
        return value

    def builtin_list_pop(value):
        if not isinstance(value, list):
            vm.runtime_error("type", "list_pop(list) expects a list")
        if not value:
            vm.runtime_error("index", "Cannot pop from an empty list")
        return value.pop()

    def from_json_value(value):
        if value is None:
            return None
        if isinstance(value, bool):
            return value
        if isinstance(value, (int, float)):
            return float(value)
        if isinstance(value, str):
            return value
        if isinstance(value, list):
            return [from_json_value(item) for item in value]
        if isinstance(value, dict):
            from nodus.vm.vm import Record
            return Record({key: from_json_value(item) for key, item in value.items()})
        vm.runtime_error("runtime", f"Unsupported JSON value: {value!r}")

    def to_json_value(value):
        from nodus.vm.vm import Record
        if value is None:
            return None
        if isinstance(value, bool):
            return value
        if isinstance(value, (int, float)) and not isinstance(value, bool):
            if isinstance(value, float) and value.is_integer():
                return int(value)
            return value
        if isinstance(value, str):
            return value
        if isinstance(value, list):
            return [to_json_value(item) for item in value]
        if isinstance(value, dict):
            return {str(key): to_json_value(item) for key, item in value.items()}
        if isinstance(value, Record):
            return {key: to_json_value(item) for key, item in value.fields.items()}
        vm.runtime_error("type", f"json.stringify cannot encode value of type {vm.builtin_type(value)}")

    def builtin_json_parse(text):
        vm.ensure_string(text, "json_parse(text)")
        try:
            parsed = json.loads(text)
        except json.JSONDecodeError as err:
            vm.runtime_error("runtime", f"json_parse failed: {err.msg}")
        return from_json_value(parsed)

    def builtin_json_stringify(value):
        try:
            return json.dumps(to_json_value(value), ensure_ascii=False)
        except Exception as err:
            from nodus.runtime.diagnostics import LangRuntimeError
            if isinstance(err, LangRuntimeError):
                raise
            vm.runtime_error("runtime", f"json_stringify failed: {err}")

    registry.add("str", 1, lambda x: vm.value_to_string(x, quote_strings=False))
    registry.add("len", 1, builtin_len)
    registry.add("collection_len", 1, builtin_len)
    registry.add("str_upper", 1, builtin_upper)
    registry.add("str_lower", 1, builtin_lower)
    registry.add("str_trim", 1, builtin_trim)
    registry.add("str_split", 2, builtin_split)
    registry.add("str_contains", 2, builtin_contains)
    registry.add("keys", 1, builtin_keys)
    registry.add("values", 1, builtin_values)
    registry.add("list_push", 2, builtin_list_push)
    registry.add("list_pop", 1, builtin_list_pop)
    registry.add("json_parse", 1, builtin_json_parse)
    registry.add("json_stringify", 1, builtin_json_stringify)



import io
import json
import os
import tempfile
import threading
import time
import unittest
from contextlib import redirect_stdout

import nodus as lang
from nodus.services.server import run_in_thread


class CliCompatibilityTests(unittest.TestCase):
    def test_cli_run_still_outputs(self):
        with tempfile.TemporaryDirectory() as td:
            script = os.path.join(td, "hello.nd")
            with open(script, "w", encoding="utf-8") as f:
                f.write('print("hello")\n')
            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "run", script])
            self.assertEqual(exit_code, 0)
            self.assertIn("hello", buf.getvalue())

    def test_cli_graph_command_outputs_plan(self):
        code = """
let A = task(fn() { return 1 }, nil)
let plan = plan_graph([A])
"""
        with tempfile.TemporaryDirectory() as td:
            script = os.path.join(td, "graph.nd")
            with open(script, "w", encoding="utf-8") as f:
                f.write(code)
            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "graph", script])
            self.assertEqual(exit_code, 0)
            payload = json.loads(buf.getvalue().strip())
            self.assertIn("graph_id", payload)
            self.assertIn("nodes", payload)

    def test_cli_test_examples_command(self):
        exit_code = lang.main(["nodus", "test-examples"])
        self.assertEqual(exit_code, 0)

    def test_cli_serve_command_starts(self):
        def run_server():
            lang.main(["nodus", "serve", "--host", "127.0.0.1", "--port", "0", "--allow-paths", "."])

        thread = threading.Thread(target=run_server, daemon=True)
        thread.start()
        time.sleep(0.05)
        self.assertTrue(thread.is_alive())

    def test_cli_snapshot_restore_worker_auth_token(self):
        token = "cli-token"
        server, thread = run_in_thread("127.0.0.1", 0, allowed_paths=["."], auth_token=token)
        port = server.server_address[1]
        try:
            time.sleep(0.05)
            os.environ["NODUS_SERVER_TOKEN"] = token
            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "snapshot", "dummy", "--host", "127.0.0.1", "--port", str(port)])
            self.assertEqual(exit_code, 1)

            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "snapshot", "dummy", "--host", "127.0.0.1", "--port", str(port), "--auth-token", token])
            self.assertEqual(exit_code, 1)

            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "snapshots", "--host", "127.0.0.1", "--port", str(port), "--auth-token", token])
            self.assertEqual(exit_code, 0)

            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "restore", "dummy", "--host", "127.0.0.1", "--port", str(port), "--auth-token", token])
            self.assertEqual(exit_code, 1)

            os.environ["NODUS_SERVER_TOKEN"] = "wrong-token"
            buf = io.StringIO()
            with redirect_stdout(buf):
                exit_code = lang.main(["nodus", "snapshots", "--host", "127.0.0.1", "--port", str(port)])
            self.assertEqual(exit_code, 1)

            os.environ["NODUS_SERVER_TOKEN"] = token
            def run_worker():
                try:
                    lang.main(["nodus", "worker", "--host", "127.0.0.1", "--port", str(port)])
                except Exception:
                    return

            worker_thread = threading.Thread(target=run_worker, daemon=True)
            worker_thread.start()
            time.sleep(0.05)
            self.assertTrue(worker_thread.is_alive())
        finally:
            os.environ.pop("NODUS_SERVER_TOKEN", None)
            server.shutdown()
            server.server_close()


if __name__ == "__main__":
    unittest.main()



"""Runtime registry and dispatch for built-in tool calls."""

from __future__ import annotations

from nodus.result import Result, normalize_filename
from nodus.vm.runtime_values import is_json_safe, payload_keys


TOOL_REGISTRY: dict[str, dict] = {}


def _ensure_registry() -> None:
    if TOOL_REGISTRY:
        return
    from nodus.tools import (
        AST_TOOL_SPEC,
        CHECK_TOOL_SPEC,
        DIS_TOOL_SPEC,
        EXECUTE_TOOL_SPEC,
        nodus_ast,
        nodus_check,
        nodus_dis,
        nodus_execute,
    )

    TOOL_REGISTRY.update(
        {
            "nodus_execute": {"handler": nodus_execute, "spec": EXECUTE_TOOL_SPEC},
            "nodus_check": {"handler": nodus_check, "spec": CHECK_TOOL_SPEC},
            "nodus_ast": {"handler": nodus_ast, "spec": AST_TOOL_SPEC},
            "nodus_dis": {"handler": nodus_dis, "spec": DIS_TOOL_SPEC},
        }
    )


def available_tools() -> list[str]:
    _ensure_registry()
    return sorted(TOOL_REGISTRY.keys())


def describe_tool(name: str):
    _ensure_registry()
    if not isinstance(name, str):
        return None
    entry = TOOL_REGISTRY.get(name)
    if entry is None:
        return None
    return dict(entry["spec"])


def call_tool(name, args, *, vm=None) -> dict:
    _ensure_registry()
    filename = normalize_filename(getattr(vm, "source_path", None))
    if not isinstance(name, str) or not name:
        return _tool_error("Tool name must be a non-empty string", filename)
    entry = TOOL_REGISTRY.get(name)
    if entry is None:
        return _tool_error(f"Unknown tool: {name}", filename, name=name)
    if not isinstance(args, dict):
        return _tool_error("tool_call(name, args) expects args as a map", filename, name=name)
    if not is_json_safe(args):
        return _tool_error("Tool args must be JSON-safe maps, lists, and primitives", filename, name=name)

    _emit(vm, "tool_call_start", name=name, payload=args)
    try:
        result = entry["handler"](**args)
    except TypeError as err:
        result = _tool_error(str(err), filename, name=name)
    except Exception as err:
        result = _tool_error(str(err), filename, name=name)

    if isinstance(result, dict) and result.get("ok"):
        _emit(vm, "tool_call_complete", name=name, payload=args, ok=True)
        return result

    _emit(vm, "tool_call_fail", name=name, payload=args, ok=False, error=_error_message(result))
    return result if isinstance(result, dict) else _tool_error("Tool returned invalid result", filename, name=name)


def _tool_error(message: str, filename: str, *, name: str | None = None) -> dict:
    legacy = {"type": "tool", "message": message, "path": filename}
    if name is not None:
        legacy["tool"] = name
    return Result.failure(
        stage="tool_call",
        filename=filename,
        stdout="",
        stderr="",
        errors=[{"type": "ToolError", "message": message, "tool": name}],
        error=legacy,
    ).to_dict()


def _error_message(result: dict | None) -> str:
    if not isinstance(result, dict):
        return "Tool call failed"
    err = result.get("error")
    if isinstance(err, dict):
        return str(err.get("message", "Tool call failed"))
    return "Tool call failed"


def _emit(vm, event_type: str, *, name: str, payload: dict, ok: bool | None = None, error: str | None = None) -> None:
    if vm is None or getattr(vm, "event_bus", None) is None:
        return
    data = {"payload_keys": payload_keys(payload)}
    if hasattr(vm, "runtime_adapter_event_data"):
        data.update(vm.runtime_adapter_event_data(payload, ok=ok, error=error))
    vm.event_bus.emit_event(event_type, name=name, data=data)



from nodus.tooling.project import LockedPackage, read_lockfile, write_lockfile


def test_lockfile_generation_is_deterministic(tmp_path):
    lock_path = tmp_path / "nodus.lock"
    packages = {
        "utils": LockedPackage(name="utils", version="1.2.3", source="path", hash="sha256:bbb", path="../utils"),
        "json": LockedPackage(name="json", version="1.2.0", source="registry", hash="sha256:aaa"),
    }

    write_lockfile(str(lock_path), packages)
    first = lock_path.read_text(encoding="utf-8")
    write_lockfile(str(lock_path), dict(reversed(list(packages.items()))))
    second = lock_path.read_text(encoding="utf-8")
    loaded = read_lockfile(str(lock_path))

    assert first == second
    assert first.startswith("[[package]]\nname = \"json\"")
    assert loaded["json"].source == "registry"
    assert loaded["utils"].path == "../utils"
    assert loaded["utils"].hash == "sha256:bbb"



"""Semantic version parsing and range matching for Nodus tooling."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True, order=True)
class Version:
    major: int
    minor: int = 0
    patch: int = 0

    @classmethod
    def parse(cls, text: str) -> "Version":
        parts = text.strip().split(".")
        if not parts or any(part == "" for part in parts):
            raise ValueError(f"Invalid version: {text}")
        values: list[int] = []
        for part in parts:
            if not part.isdigit():
                raise ValueError(f"Invalid version: {text}")
            values.append(int(part))
        while len(values) < 3:
            values.append(0)
        if len(values) > 3:
            raise ValueError(f"Invalid version: {text}")
        return cls(values[0], values[1], values[2])

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"


@dataclass(frozen=True)
class Comparator:
    op: str
    version: Version

    def matches(self, value: Version) -> bool:
        if self.op == "==":
            return value == self.version
        if self.op == ">":
            return value > self.version
        if self.op == ">=":
            return value >= self.version
        if self.op == "<":
            return value < self.version
        if self.op == "<=":
            return value <= self.version
        raise ValueError(f"Unsupported operator: {self.op}")


class VersionRange:
    def __init__(self, comparators: list[Comparator]):
        self.comparators = comparators

    @classmethod
    def parse(cls, text: str) -> "VersionRange":
        raw = text.strip()
        if raw.startswith("^"):
            base = Version.parse(raw[1:])
            return cls([Comparator(">=", base), Comparator("<", _caret_upper(base))])
        if raw.startswith("~"):
            base = Version.parse(raw[1:])
            return cls([Comparator(">=", base), Comparator("<", Version(base.major, base.minor + 1, 0))])

        parts = [part.strip() for part in raw.split(",") if part.strip()]
        if not parts:
            raise ValueError(f"Invalid range: {text}")
        comparators: list[Comparator] = []
        for part in parts:
            op = None
            for candidate in (">=", "<=", ">", "<", "=="):
                if part.startswith(candidate):
                    op = candidate
                    comparators.append(Comparator(candidate, Version.parse(part[len(candidate) :].strip())))
                    break
            if op is None:
                comparators.append(Comparator("==", Version.parse(part)))
        return cls(comparators)

    def matches(self, value: Version) -> bool:
        return all(comp.matches(value) for comp in self.comparators)


def _caret_upper(base: Version) -> Version:
    if base.major > 0:
        return Version(base.major + 1, 0, 0)
    if base.minor > 0:
        return Version(0, base.minor + 1, 0)
    return Version(0, 0, base.patch + 1)



"""Compatibility wrappers for tooling-side project helpers."""

from nodus.tooling.installer import install_project
from nodus.tooling.project import (
    DependencySpec,
    LockedPackage as ResolvedDependency,
    ProjectConfig,
    create_project,
    find_project_root,
    load_manifest,
    load_project,
    load_project_from,
    parse_dependencies,
    read_lockfile,
    write_lockfile,
)
from nodus.tooling.registry import Registry
from nodus.tooling.resolver import ResolutionResult, resolve_project_dependencies


def resolve_dependencies(project: ProjectConfig, *, update: bool = False) -> ResolutionResult:
    return resolve_project_dependencies(project, update=update, registry=Registry.from_project_root(project.root))


def install_dependencies(project: ProjectConfig, *, update: bool = False) -> dict[str, ResolvedDependency]:
    resolution = resolve_dependencies(project, update=update)
    return install_project(project, resolution)



"""Coroutine scheduler and event loop for Nodus."""

from __future__ import annotations

TASK_STEP_BUDGET = 1000

import heapq
import sys
import time
from collections import deque

from nodus.runtime.diagnostics import LangRuntimeError, format_error
from nodus.runtime.runtime_stats import runtime_time_ms
from nodus.runtime.runtime_events import RuntimeEvent

SLEEP_KEY = "__sleep_ms"
CHANNEL_WAIT_KEY = "__channel_wait"


class SleepRequest:
    def __init__(self, ms: float):
        self.ms = ms


class Scheduler:
    def __init__(self, vm, *, trace: bool = False, trace_output=print):
        self.vm = vm
        self.ready_queue = deque()
        self.queue = self.ready_queue
        self.timers: list[tuple[float, int, object]] = []
        self.sleeping_tasks: set[int] = set()
        self.completed_tasks: list[object] = []
        self._completed_ids: set[int] = set()
        self.total_tasks_spawned = 0
        self.total_resumes = 0
        self.tasks: dict[int, object] = {}
        self.current_task: object | None = None
        self._next_id = 1
        self.trace = trace
        self.trace_output = trace_output
        self._counter = 0
        self.task_ages: dict[int, int] = {}

    def _trace(self, message: str) -> None:
        if self.trace:
            self.trace_output(message)

    def _emit_event(self, event_type: str, coroutine=None, data: dict | None = None) -> None:
        if not hasattr(self.vm, "event_bus") or self.vm.event_bus is None:
            return
        coroutine_id = None
        name = None
        if coroutine is not None:
            self._ensure_metadata(coroutine)
            coroutine_id = coroutine.id
            name = coroutine.name
        self.vm.event_bus.emit(RuntimeEvent(event_type, runtime_time_ms(), coroutine_id=coroutine_id, name=name, data=data))

    def _ensure_metadata(self, coroutine) -> None:
        if coroutine.id is None:
            coroutine.id = self._next_id
            self._next_id += 1
            self.total_tasks_spawned += 1
            self.tasks[coroutine.id] = coroutine
            self.task_ages[coroutine.id] = self.total_tasks_spawned
        if coroutine.name is None and getattr(coroutine, "closure", None) is not None:
            coroutine.name = coroutine.closure.function.display_name
        if coroutine.name is None:
            coroutine.name = "<anonymous>"
        if coroutine.module is None and getattr(coroutine, "closure", None) is not None:
            module_path, _line, _col = self.vm.code_locs[coroutine.closure.function.addr]
            coroutine.module = module_path or self.vm.source_path
        if coroutine.created_time is None:
            coroutine.created_time = runtime_time_ms()

    def run_task_graph(self, graph) -> object:
        return self.vm.builtin_run_graph(graph)

    def spawn(self, coroutine) -> None:
        if coroutine.state == "finished":
            return
        self._ensure_metadata(coroutine)
        self.ready_queue.append(coroutine)
        self._emit_event("coroutine_spawn", coroutine)
        self._trace(f"spawn coroutine #{coroutine.id} {coroutine.name}")
        if coroutine.id is not None:
            self.task_ages[coroutine.id] = self.total_tasks_spawned

    def schedule(self, coroutine) -> None:
        if coroutine.state == "finished":
            return
        self.ready_queue.append(coroutine)
        if coroutine.id is not None:
            self.task_ages[coroutine.id] = self.total_resumes

    def _schedule_sleep(self, coroutine, ms: float) -> None:
        delay = max(0.0, ms) / 1000.0
        self._counter += 1
        heapq.heappush(self.timers, (runtime_time_ms() + delay * 1000.0, self._counter, coroutine))
        if coroutine.id is not None:
            self.sleeping_tasks.add(coroutine.id)

    def schedule_delay(self, coroutine, ms: float) -> None:
        if coroutine.state == "finished":
            return
        self._ensure_metadata(coroutine)
        self._schedule_sleep(coroutine, ms)

    def _drain_timers(self) -> None:
        now = runtime_time_ms()
        while self.timers and self.timers[0][0] <= now:
            _wake, _seq, coroutine = heapq.heappop(self.timers)
            if coroutine.state != "finished":
                if coroutine.id is not None:
                    self.sleeping_tasks.discard(coroutine.id)
                self.ready_queue.append(coroutine)
                self._emit_event("coroutine_wake", coroutine)

    def _extract_sleep_ms(self, value):
        if isinstance(value, dict) and SLEEP_KEY in value:
            return value[SLEEP_KEY]
        return None

    def _extract_channel_wait(self, value):
        if isinstance(value, dict) and CHANNEL_WAIT_KEY in value:
            return value[CHANNEL_WAIT_KEY]
        return None

    def _mark_completed(self, coroutine) -> None:
        if coroutine.id is None:
            return
        if coroutine.id in self._completed_ids:
            return
        self._completed_ids.add(coroutine.id)
        self.completed_tasks.append(coroutine)

    def run_loop(self, on_complete=None, on_error=None) -> None:
        stop = False
        while self.ready_queue or self.timers:
            self._drain_timers()
            if not self.ready_queue:
                if not self.timers:
                    break
                wake_time = self.timers[0][0]
                now = runtime_time_ms()
                if wake_time > now:
                    time.sleep((wake_time - now) / 1000.0)
                self._drain_timers()
                if not self.ready_queue:
                    continue

            coroutine = self.ready_queue.popleft()
            if coroutine.state == "finished":
                continue
            if coroutine.task_timeout_ms is not None and coroutine.task_started_at is not None:
                now = runtime_time_ms()
                if now - coroutine.task_started_at > coroutine.task_timeout_ms:
                    err = LangRuntimeError("timeout", "Task timed out")
                    self._mark_completed(coroutine)
                    if coroutine.id is not None:
                        self.sleeping_tasks.discard(coroutine.id)
                    if on_error is not None:
                        stop = bool(on_error(coroutine, err))
                    if stop:
                        break
                    continue
            try:
                self.current_task = coroutine
                self.total_resumes += 1
                coroutine.resume_count += 1
                now = runtime_time_ms()
                coroutine.last_resume = now
                coroutine.last_run_time = now
                self._emit_event("coroutine_resume", coroutine)
                self._trace(f"resume coroutine #{coroutine.id}")
                self.vm.task_step_budget = TASK_STEP_BUDGET
                self.vm._budget_exceeded = False
                result = self.vm.builtin_coroutine_resume(coroutine)
            except Exception as err:
                print(format_error(err, path=self.vm.source_path), file=sys.stderr)
                self._mark_completed(coroutine)
                if coroutine.id is not None:
                    self.sleeping_tasks.discard(coroutine.id)
                if on_error is not None:
                    stop = bool(on_error(coroutine, err))
                if stop:
                    break
                continue
            finally:
                self.current_task = None
                self.vm.task_step_budget = None
                self.vm._budget_exceeded = False

            if coroutine.state != "suspended":
                if coroutine.state == "finished":
                    self._mark_completed(coroutine)
                    self._emit_event("coroutine_complete", coroutine)
                    self._trace(f"complete coroutine #{coroutine.id}")
                    if on_complete is not None:
                        stop = bool(on_complete(coroutine)) or stop
                continue

            channel_wait = self._extract_channel_wait(result)
            if channel_wait is not None:
                self._trace(f"block coroutine #{coroutine.id} channel")
                continue

            sleep_ms = self._extract_sleep_ms(result)
            if sleep_ms is not None:
                if isinstance(sleep_ms, bool) or not isinstance(sleep_ms, (int, float)):
                    err = LangRuntimeError("type", "sleep(ms) expects a number")
                    if hasattr(self.vm, "emit_runtime_error"):
                        self.vm.emit_runtime_error(err)
                    print(format_error(err, path=self.vm.source_path), file=sys.stderr)
                    continue
                self._trace(f"sleep coroutine #{coroutine.id} {float(sleep_ms)}")
                self._emit_event("coroutine_sleep", coroutine, {"ms": float(sleep_ms)})
                self._schedule_sleep(coroutine, float(sleep_ms))
                continue

            self._trace(f"yield coroutine #{coroutine.id}")
            self._emit_event("coroutine_yield", coroutine)
            self.ready_queue.append(coroutine)
            if stop:
                break



import os

from nodus.tooling.runner import run_source


def _write(tmp_path, name: str, content: str) -> str:
    path = tmp_path / name
    path.write_text(content, encoding="utf-8")
    return str(path)


def test_module_import_execution(tmp_path):
    util_path = _write(
        tmp_path,
        "util.nd",
        'export let value = 3\n',
    )
    main_path = _write(
        tmp_path,
        "main.nd",
        f'import {{ value }} from "{os.path.basename(util_path)}"\nprint(value)\n',
    )
    code = open(main_path, "r", encoding="utf-8").read()
    result, _vm = run_source(code, filename=main_path)
    assert result["ok"] is True
    assert result["stdout"].strip() == "3.0"


def test_module_caching(tmp_path):
    counter_path = _write(
        tmp_path,
        "counter.nd",
        "export let counter = 0\ncounter = counter + 1\n",
    )
    main_path = _write(
        tmp_path,
        "main.nd",
        f'import "{os.path.basename(counter_path)}" as a\nimport "{os.path.basename(counter_path)}" as b\nprint(a.counter)\nprint(b.counter)\n',
    )
    code = open(main_path, "r", encoding="utf-8").read()
    result, _vm = run_source(code, filename=main_path)
    assert result["ok"] is True
    assert result["stdout"].strip() == "1.0\n1.0"


def test_export_visibility(tmp_path):
    mod_path = _write(
        tmp_path,
        "mod.nd",
        "let hidden = 5\nexport let visible = 1\n",
    )
    main_path = _write(
        tmp_path,
        "main.nd",
        f'import {{ hidden }} from "{os.path.basename(mod_path)}"\n',
    )
    code = open(main_path, "r", encoding="utf-8").read()
    result, _vm = run_source(code, filename=main_path)
    assert result["ok"] is False
    assert result["error"]["type"] == "runtime"
    assert result["error"]["kind"] == "import"


def test_module_loader_resolves_installed_modules_before_stdlib(tmp_path):
    (tmp_path / "nodus.toml").write_text('name = "demo"\nversion = "0.1.0"\n', encoding="utf-8")
    dep_dir = tmp_path / ".nodus" / "modules" / "utils"
    dep_dir.mkdir(parents=True)
    (dep_dir / "strings.nd").write_text('export fn upper(value) { return "dep:" + value }\n', encoding="utf-8")
    main_path = _write(
        tmp_path,
        "main.nd",
        'import "utils:strings" as strings\nprint(strings.upper("ok"))\n',
    )
    code = open(main_path, "r", encoding="utf-8").read()
    result, _vm = run_source(code, filename=main_path)
    assert result["ok"] is True
    assert result["stdout"].strip() == "dep:ok"



import io
import unittest
from contextlib import redirect_stdout

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader


def run_program(src: str, source_path: str | None = None) -> list[str]:
    _loader = ModuleLoader(project_root=None)
    code, functions, code_locs = _loader.compile_only(
        src,
        module_name=source_path or "<memory>",
    )
    vm = lang.VM(code, functions, code_locs=code_locs, source_path=source_path)
    buf = io.StringIO()
    with redirect_stdout(buf):
        vm.run()
    return buf.getvalue().splitlines()


class RecordTests(unittest.TestCase):
    def test_basic_record(self):
        src = """
let user = record {
    name: "Alice",
    age: 25
}
print(user.name)
"""
        self.assertEqual(run_program(src), ["Alice"])

    def test_field_mutation(self):
        src = """
let user = record {
    age: 20
}
user.age = 21
print(user.age)
"""
        self.assertEqual(run_program(src), ["21.0"])

    def test_nested_records(self):
        src = """
let user = record {
    profile: record {
        name: "Bob"
    }
}
print(user.profile.name)
"""
        self.assertEqual(run_program(src), ["Bob"])

    def test_missing_field(self):
        src = """
let user = record {}
print(user.name)
"""
        with self.assertRaises(lang.LangRuntimeError):
            run_program(src, source_path="main.nd")


if __name__ == "__main__":
    unittest.main()



"""Registry metadata access for tooling-side dependency resolution."""

from __future__ import annotations

import os
from dataclasses import dataclass, field

from nodus.tooling.project import NODUS_DIRNAME, load_manifest


REGISTRY_NAME = "registry.toml"


@dataclass(frozen=True)
class RegistryPackage:
    name: str
    version: str
    source: str
    path: str
    dependencies: dict[str, str] = field(default_factory=dict)


class Registry:
    def __init__(self, packages: dict[str, dict[str, RegistryPackage]] | None = None) -> None:
        self._packages = packages or {}

    @classmethod
    def from_project_root(cls, project_root: str) -> "Registry":
        path = os.path.join(project_root, NODUS_DIRNAME, REGISTRY_NAME)
        if not os.path.isfile(path):
            return cls()
        data = load_manifest(path)
        raw_packages = data.get("packages", {})
        packages: dict[str, dict[str, RegistryPackage]] = {}
        if not isinstance(raw_packages, dict):
            raise ValueError("Registry packages must be a table")
        for raw_name, versions in raw_packages.items():
            name = str(raw_name)
            if not isinstance(versions, dict):
                raise ValueError(f"Registry package {name} must define versions as a table")
            package_versions: dict[str, RegistryPackage] = {}
            for raw_version, entry in versions.items():
                version = str(raw_version)
                if not isinstance(entry, dict):
                    raise ValueError(f"Registry package {name}@{version} must be a table")
                raw_path = entry.get("path")
                if raw_path is None:
                    raise ValueError(f"Registry package {name}@{version} is missing path")
                dependencies = entry.get("dependencies", {})
                if not isinstance(dependencies, dict):
                    raise ValueError(f"Registry package {name}@{version} dependencies must be a table")
                package_versions[version] = RegistryPackage(
                    name=name,
                    version=version,
                    source=str(entry.get("source", "registry")),
                    path=_resolve_registry_path(path, str(raw_path)),
                    dependencies={str(dep_name): str(dep_spec) for dep_name, dep_spec in dependencies.items()},
                )
            packages[name] = package_versions
        return cls(packages)

    def available_versions(self, name: str) -> list[RegistryPackage]:
        return [self._packages[name][version] for version in sorted(self._packages.get(name, {}), key=_version_sort_key)]

    def get(self, name: str, version: str) -> RegistryPackage | None:
        return self._packages.get(name, {}).get(version)


def _resolve_registry_path(registry_path: str, raw_path: str) -> str:
    if os.path.isabs(raw_path):
        return os.path.abspath(raw_path)
    return os.path.abspath(os.path.join(os.path.dirname(registry_path), raw_path))


def _version_sort_key(text: str) -> tuple[int, int, int, str]:
    parts = text.split(".")
    numeric = []
    for part in parts[:3]:
        numeric.append(int(part) if part.isdigit() else -1)
    while len(numeric) < 3:
        numeric.append(0)
    return numeric[0], numeric[1], numeric[2], text



import os

import nodus as lang

from nodus.runtime import bytecode_cache
from nodus.runtime.module_loader import ModuleLoader


def _write(tmp_path, name: str, content: str) -> str:
    path = tmp_path / name
    path.write_text(content, encoding="utf-8")
    return str(path)


def _bump_mtime(path: str) -> None:
    current = os.stat(path).st_mtime_ns
    updated = current + 2_000_000_000
    os.utime(path, ns=(updated, updated))


def test_bytecode_cache_written_after_first_compile(tmp_path):
    main_path = _write(tmp_path, "main.nd", "let value = 1\n")

    loader = ModuleLoader(project_root=str(tmp_path))
    loader.load_module_from_path(main_path)

    cache_root = tmp_path / ".nodus" / "cache"
    files = sorted(cache_root.glob("*.nbc"))
    assert len(files) == 1
    cached = bytecode_cache.load_cached_bytecode(str(tmp_path), main_path)
    assert cached is not None
    assert cached.code.get("module_name") == os.path.abspath(main_path)


def test_bytecode_cache_reused_on_second_run(tmp_path, monkeypatch):
    main_path = _write(tmp_path, "main.nd", "let value = 1\n")
    ModuleLoader(project_root=str(tmp_path)).load_module_from_path(main_path)

    compile_calls = []
    original = ModuleLoader._compile_module

    def counting_compile(self, metadata):
        compile_calls.append(metadata.module_id)
        return original(self, metadata)

    monkeypatch.setattr(ModuleLoader, "_compile_module", counting_compile)

    loader = ModuleLoader(project_root=str(tmp_path))
    loader.load_module_from_path(main_path)

    assert compile_calls == []


def test_bytecode_cache_invalidates_when_source_changes(tmp_path, monkeypatch):
    main_path = _write(tmp_path, "main.nd", "let value = 1\n")
    ModuleLoader(project_root=str(tmp_path)).load_module_from_path(main_path)

    with open(main_path, "w", encoding="utf-8") as handle:
        handle.write("let value = 2\n")
    _bump_mtime(main_path)

    compile_calls = []
    original = ModuleLoader._compile_module

    def counting_compile(self, metadata):
        compile_calls.append(metadata.module_id)
        return original(self, metadata)

    monkeypatch.setattr(ModuleLoader, "_compile_module", counting_compile)

    loader = ModuleLoader(project_root=str(tmp_path))
    loader.load_module_from_path(main_path)

    assert compile_calls == [os.path.abspath(main_path)]


def test_bytecode_cache_invalidates_when_version_changes(tmp_path, monkeypatch):
    main_path = _write(tmp_path, "main.nd", "let value = 1\n")
    ModuleLoader(project_root=str(tmp_path)).load_module_from_path(main_path)

    compile_calls = []
    original = ModuleLoader._compile_module

    def counting_compile(self, metadata):
        compile_calls.append(metadata.module_id)
        return original(self, metadata)

    monkeypatch.setattr(ModuleLoader, "_compile_module", counting_compile)
    monkeypatch.setattr(bytecode_cache, "NODUS_BYTECODE_VERSION", bytecode_cache.NODUS_BYTECODE_VERSION + 1)

    loader = ModuleLoader(project_root=str(tmp_path))
    loader.load_module_from_path(main_path)

    assert compile_calls == [os.path.abspath(main_path)]


def test_cli_cache_clear_removes_cached_bytecode(tmp_path):
    (tmp_path / "nodus.toml").write_text('name = "demo"\nversion = "0.1.0"\n', encoding="utf-8")
    main_path = _write(tmp_path, "main.nd", "let value = 1\n")
    ModuleLoader(project_root=str(tmp_path)).load_module_from_path(main_path)

    cache_root = tmp_path / ".nodus" / "cache"
    assert any(cache_root.iterdir())

    exit_code = lang.main(["nodus", "cache", "clear", "--path", str(tmp_path)])

    assert exit_code == 0
    assert list(cache_root.iterdir()) == []



from nodus.tooling.project import load_project
from nodus.tooling.registry import Registry
from nodus.tooling.resolver import resolve_project_dependencies


def test_dependency_graph_resolution_with_path_and_registry_sources(tmp_path):
    utils = tmp_path / "utils"
    utils.mkdir()
    (utils / "nodus.toml").write_text('name = "utils"\nversion = "1.2.3"\n', encoding="utf-8")
    (utils / "util.nd").write_text("export let value = 1\n", encoding="utf-8")

    registry_root = tmp_path / "registry"
    json_v1 = registry_root / "json-1.1.0"
    json_v2 = registry_root / "json-1.2.0"
    json_v1.mkdir(parents=True)
    json_v2.mkdir(parents=True)
    (json_v1 / "nodus.toml").write_text('name = "json"\nversion = "1.1.0"\n', encoding="utf-8")
    (json_v2 / "nodus.toml").write_text('name = "json"\nversion = "1.2.0"\n', encoding="utf-8")
    (json_v1 / "json.nd").write_text('export let version = "1.1.0"\n', encoding="utf-8")
    (json_v2 / "json.nd").write_text('export let version = "1.2.0"\n', encoding="utf-8")

    project = tmp_path / "app"
    project.mkdir()
    (project / ".nodus").mkdir()
    (project / "nodus.toml").write_text(
        '\n'.join(
            [
                'name = "app"',
                'version = "0.1.0"',
                "",
                "[dependencies]",
                'json = "^1.1.0"',
                'utils = { path = "../utils" }',
            ]
        )
        + "\n",
        encoding="utf-8",
    )
    (project / ".nodus" / "registry.toml").write_text(
        '\n'.join(
            [
                '[packages.json."1.1.0"]',
                'path = "../../registry/json-1.1.0"',
                "",
                '[packages.json."1.2.0"]',
                'path = "../../registry/json-1.2.0"',
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    config = load_project(str(project))
    resolution = resolve_project_dependencies(config, update=True, registry=Registry.from_project_root(config.root))

    assert resolution.packages["utils"].version == "1.2.3"
    assert resolution.packages["json"].version == "1.2.0"
    assert resolution.graph["utils"] == ()
    assert resolution.graph["json"] == ()
    assert set(resolution.install_order) == {"json", "utils"}



import io
import unittest
from contextlib import redirect_stdout

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader


def run_program(src: str, source_path: str | None = None) -> list[str]:
    _loader = ModuleLoader(project_root=None)
    code, functions, code_locs = _loader.compile_only(
        src,
        module_name=source_path or "<memory>",
    )
    vm = lang.VM(code, functions, code_locs=code_locs, source_path=source_path)
    buf = io.StringIO()
    with redirect_stdout(buf):
        vm.run()
    return buf.getvalue().splitlines()


class ClosureTests(unittest.TestCase):
    def test_basic_closure(self):
        src = """
fn outer() {
    let x = 10
    fn inner() {
        return x
    }
    return inner
}
let f = outer()
print(f())
"""
        self.assertEqual(run_program(src), ["10.0"])

    def test_closure_mutation(self):
        src = """
fn counter() {
    let x = 0
    fn inc() {
        x = x + 1
        return x
    }
    return inc
}
let c = counter()
print(c())
print(c())
"""
        self.assertEqual(run_program(src), ["1.0", "2.0"])

    def test_nested_closure(self):
        src = """
fn a() {
    let x = 1
    fn b() {
        fn c() {
            return x
        }
        return c
    }
    return b()
}
print(a()())
"""
        self.assertEqual(run_program(src), ["1.0"])


if __name__ == "__main__":
    unittest.main()



"""Tests for formatter coverage of previously-missing AST node handlers."""
import unittest

from nodus.tooling.formatter import format_source


class FormatterCoverageTests(unittest.TestCase):
    def test_yield_no_expr(self):
        src = "fn f() {\n    yield\n}\n"
        self.assertEqual(format_source(src), src)

    def test_yield_with_expr(self):
        src = "fn f() {\n    yield 42\n}\n"
        self.assertEqual(format_source(src), src)

    def test_throw(self):
        src = 'fn f() {\n    throw "error"\n}\n'
        self.assertEqual(format_source(src), src)

    def test_try_catch(self):
        src = 'fn f() {\n    try {\n        throw "oops"\n    } catch err {\n        print(err)\n    }\n}\n'
        self.assertEqual(format_source(src), src)

    def test_destructure_list(self):
        src = "let [a, b] = xs\n"
        self.assertEqual(format_source(src), src)

    def test_destructure_record(self):
        src = "let {x: a, y: b} = pt\n"
        self.assertEqual(format_source(src), src)

    def test_destructure_nested(self):
        src = "let [a, [b, c]] = xs\n"
        self.assertEqual(format_source(src), src)


if __name__ == "__main__":
    unittest.main()



import unittest

from nodus.tooling.runner import build_ast


class AstSerializationTests(unittest.TestCase):
    def test_ast_dict_structure(self):
        result = build_ast("let x = 1", filename="inline.nd")
        self.assertTrue(result["ok"])
        self.assertIn("ast", result)
        ast = result["ast"]
        self.assertIsInstance(ast, list)
        self.assertTrue(ast)
        let_node = ast[0]
        self.assertEqual(let_node["type"], "Let")
        self.assertEqual(let_node["name"], "x")
        expr = let_node["expr"]
        self.assertEqual(expr["type"], "Num")
        self.assertEqual(expr["v"], 1.0)


if __name__ == "__main__":
    unittest.main()



"""Execution sessions for the Nodus runtime server."""

from __future__ import annotations

import secrets
import time
from dataclasses import dataclass

from nodus.support.config import MAX_SESSIONS, SESSION_TIMEOUT_MS
from nodus.services.memory_runtime import GLOBAL_MEMORY_STORE, MemoryStore
from nodus.vm.vm import VM


@dataclass
class Session:
    id: str
    vm: VM
    created_time: float
    last_access: float
    execution_count: int = 0
    import_state: dict | None = None


class SessionManager:
    def __init__(self, *, timeout_ms: int = SESSION_TIMEOUT_MS, max_sessions: int = MAX_SESSIONS):
        self.timeout_ms = timeout_ms
        self.max_sessions = max_sessions
        self.sessions: dict[str, Session] = {}

    def _is_expired(self, last_access: float, now: float) -> bool:
        if self.timeout_ms is None:
            return False
        grace_ms = min(10.0, self.timeout_ms * 0.2)
        return (now - last_access) * 1000.0 > (self.timeout_ms + grace_ms)

    def create(self, vm: VM) -> Session:
        self.cleanup()
        if len(self.sessions) >= self.max_sessions:
            self._evict_oldest()
        session_id = secrets.token_hex(16)
        now = time.monotonic()
        session = Session(
            id=session_id,
            vm=vm,
            created_time=now,
            last_access=now,
            execution_count=0,
            import_state={"loaded": set(), "loading": set(), "exports": {}, "modules": {}, "module_ids": {}, "project_root": None},
        )
        vm.session_id = session_id
        if not isinstance(getattr(vm, "memory_store", None), MemoryStore) or vm.memory_store is GLOBAL_MEMORY_STORE:
            vm.memory_store = MemoryStore()
        self.sessions[session_id] = session
        return session

    def create_from_state(self, state: dict, import_state: dict | None = None) -> Session:
        vm = VM([], {}, code_locs=[], source_path=None)
        vm.memory_store = MemoryStore()
        vm.import_state(state)
        session = self.create(vm)
        session.import_state = import_state or {"loaded": set(), "loading": set(), "exports": {}, "modules": {}, "module_ids": {}, "project_root": None}
        return session

    def get(self, session_id: str) -> Session | None:
        self.cleanup()
        return self.sessions.get(session_id)

    def touch(self, session: Session) -> None:
        session.last_access = time.monotonic()

    def record_execution(self, session: Session) -> None:
        session.execution_count += 1
        self.touch(session)

    def cleanup(self) -> None:
        if self.timeout_ms is None:
            return
        now = time.monotonic()
        expired = []
        for session_id, session in self.sessions.items():
            if self._is_expired(session.last_access, now):
                expired.append(session_id)
        for session_id in expired:
            self.sessions.pop(session_id, None)

    def list_sessions(self) -> list[dict]:
        self.cleanup()
        out = []
        for session in self.sessions.values():
            out.append(
                {
                    "id": session.id,
                    "created_time": session.created_time,
                    "last_access": session.last_access,
                    "execution_count": session.execution_count,
                }
            )
        return out

    def _evict_oldest(self) -> None:
        if not self.sessions:
            return
        oldest_id = min(self.sessions.items(), key=lambda item: item[1].last_access)[0]
        self.sessions.pop(oldest_id, None)



import io
import unittest
from contextlib import redirect_stderr, redirect_stdout

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader
from nodus.tooling.runner import run_in_vm
import json
import os
from nodus.services.server import run_in_thread
import threading
import http.client
import time
from nodus.orchestration.task_graph import set_default_dispatcher
from nodus.runtime.runtime_events import RuntimeEventBus


def run_program(src: str, source_path: str | None = None):
    _loader = ModuleLoader(project_root=None)
    code, functions, code_locs = _loader.compile_only(
        src,
        module_name=source_path or "<memory>",
    )
    vm = lang.VM(code, functions, code_locs=code_locs, source_path=source_path)
    out_buf = io.StringIO()
    err_buf = io.StringIO()
    with redirect_stdout(out_buf), redirect_stderr(err_buf):
        vm.run()
    return vm, out_buf.getvalue().splitlines(), err_buf.getvalue()


class TaskGraphTests(unittest.TestCase):
    def _poll_job(self, worker_manager, worker_id: str, timeout: float = 10.0):
        return worker_manager.wait_for_job(worker_id, timeout=timeout)

    def test_dependency_ordering(self):
        src = """
let A = task(fn() { return 2 }, nil)
let B = task(fn(x) { return x * 3 }, A)
let C = task(fn(x) { return x + 1 }, B)

let result = run_graph([A, B, C])
print(result["tasks"]["task_1"])
print(result["tasks"]["task_2"])
print(result["tasks"]["task_3"])
"""
        _vm, out, err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["2.0", "6.0", "7.0"])
        self.assertEqual(err.strip(), "")

    def test_parallel_execution(self):
        src = """
let A = task(fn() { return 2 }, nil)
let B = task(fn() { return 3 }, nil)
let C = task(fn(x, y) { return x + y }, [A, B])

run_graph([A, B, C])
"""
        vm, _out, _err = run_program(src, source_path="main.nd")
        starts = [e.name for e in vm.event_bus.events() if e.type == "task_start"]
        self.assertEqual(starts[0:2], ["task_1", "task_2"])
        self.assertEqual(starts[2], "task_3")

    def test_error_propagation(self):
        src = """
let A = task(fn() { return 2 }, nil)
let B = task(fn(x) { throw "boom" }, A)
let C = task(fn(x) { return x + 1 }, B)

let result = run_graph([A, B, C])
print(result["failed"])
"""
        _vm, out, _err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["[\"task_2\"]"])

    def test_task_yield(self):
        src = """
let A = task(fn() {
    yield 0
    return 2
}, nil)
let B = task(fn(x) { return x + 1 }, A)
let result = run_graph([A, B])
print(result["tasks"]["task_2"])
"""
        _vm, out, err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["3.0"])
        self.assertEqual(err.strip(), "")

    def test_task_timeout(self):
        src = """
let A = task(fn() {
    while (true) { yield 0 }
    return 1
}, { "deps": nil, "timeout_ms": 10 })
let result = run_graph([A])
print(result["failed"])
"""
        _vm, out, _err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["[\"task_1\"]"])

    def test_task_retry_success(self):
        src = """
let state = { "count": 0 }
let A = task(fn() {
    if (state["count"] == 0) {
        state["count"] = 1
        throw "fail"
    }
    return 5
}, { "retries": 2, "retry_delay_ms": 5 })
let result = run_graph([A])
print(result["tasks"]["task_1"])
print(result["attempts"]["task_1"])
"""
        _vm, out, _err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["5.0", "2.0"])

    def test_task_retry_exhausted(self):
        src = """
let A = task(fn() {
    throw "fail"
}, { "retries": 1, "retry_delay_ms": 1 })
let result = run_graph([A])
print(result["failed"])
print(result["attempts"]["task_1"])
"""
        _vm, out, _err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["[\"task_1\"]", "2.0"])

    def test_task_retry_delay(self):
        src = """
let state = { "count": 0 }
let A = task(fn() {
    if (state["count"] == 0) {
        state["count"] = 1
        throw "fail"
    }
    return 5
}, { "retries": 1, "retry_delay_ms": 50 })
let result = run_graph([A])
let timing = result["timings"]["task_1"]
print(timing["finished_at"] - timing["started_at"])
"""
        _vm, out, _err = run_program(src, source_path="main.nd")
        self.assertTrue(float(out[0]) >= 50.0)

    def test_task_cache_hit(self):
        src = """
let A = task(fn() { return 5 }, { "cache": true })
let result1 = run_graph([A])
let result2 = run_graph([A])
print(result1["tasks"]["task_1"])
print(result2["cache_hits"][0])
"""
        _vm, out, _err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["5.0", "task_1"])

    def test_task_cache_disabled(self):
        src = """
let state = { "count": 0 }
let A = task(fn() { state["count"] = state["count"] + 1 return state["count"] }, nil)
let result1 = run_graph([A])
let result2 = run_graph([A])
print(result1["tasks"]["task_1"])
print(result2["tasks"]["task_1"])
"""
        _vm, out, _err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["1.0", "2.0"])

    def test_task_cache_with_dependencies(self):
        src = """
let A = task(fn() { return 2 }, { "cache": true })
let B = task(fn(x) { return x + 3 }, { "deps": A, "cache": true })
let result1 = run_graph([A, B])
let result2 = run_graph([A, B])
print(result1["tasks"]["task_2"])
print(result2["cache_hits"][0])
"""
        _vm, out, _err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["5.0", "task_1"])

    def test_graph_plan_linear(self):
        src = """
let A = task(fn() { return 2 }, nil)
let B = task(fn(x) { return x * 3 }, A)
let C = task(fn(x) { return x + 1 }, B)
let plan = plan_graph([A, B, C])
print(len(plan["levels"]))
"""
        _vm, out, _err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["3.0"])

    def test_graph_plan_branching(self):
        src = """
let A = task(fn() { return 1 }, nil)
let B = task(fn(x) { return x + 1 }, A)
let C = task(fn(x) { return x * 2 }, A)
let plan = plan_graph([A, B, C])
print(len(plan["parallel_groups"][1]))
"""
        _vm, out, _err = run_program(src, source_path="main.nd")
        self.assertEqual(out, ["2.0"])

    def test_graph_persistence_and_resume(self):
        src = """
let state = { "count": 0 }
let A = task(fn() { state["count"] = state["count"] + 1 return 2 }, nil)
let B = task(fn(x) { return x + 1 }, A)
let plan = plan_graph([A, B])
"""
        _loader = ModuleLoader(project_root=None)
        code, functions, code_locs = _loader.compile_only(src, module_name="main.nd")
        vm = lang.VM(code, functions, code_locs=code_locs, source_path="main.nd")
        vm.run()
        graph_id = None
        if getattr(vm, "last_graph_plan", None):
            graph_id = vm.last_graph_plan.get("graph_id")
        if graph_id is None:
            plan_val = vm.globals.get("plan")
            if hasattr(plan_val, "value"):
                plan_val = plan_val.value
            graph_id = plan_val.get("graph_id") if isinstance(plan_val, dict) else None
        if graph_id is None:
            graph_dir = os.path.join(".nodus", "graphs")
            if os.path.isdir(graph_dir):
                files = [f for f in os.listdir(graph_dir) if f.endswith(".json")]
                if files:
                    files.sort()
                    graph_id = files[-1].rsplit(".", 1)[0]
        self.assertIsNotNone(graph_id)
        path = os.path.join(".nodus", "graphs", f"{graph_id}.json")
        os.makedirs(os.path.dirname(path), exist_ok=True)
        state = {
            "graph_id": graph_id,
            "status": "running",
            "tasks": {
                "task_1": {"state": "completed", "result": 2, "attempts": 1},
                "task_2": {"state": "pending", "attempts": 0},
            },
        }
        with open(path, "w", encoding="utf-8") as f:
            json.dump(state, f)
        result = vm.builtin_resume_graph(graph_id)
        self.assertEqual(result["tasks"]["task_2"], 3)

    def test_worker_execution(self):
        from nodus.services.server import WorkerManager
        worker_manager = WorkerManager()
        set_default_dispatcher(worker_manager)
        worker_id = worker_manager.register()

        src = """
let A = task(fn() { return 5 }, nil)
let result = run_graph([A])
print(result["tasks"]["task_1"])
"""
        _loader = ModuleLoader(project_root=None)
        code, functions, code_locs = _loader.compile_only(src, module_name="main.nd")
        vm = lang.VM(code, functions, code_locs=code_locs, source_path="main.nd")
        vm.worker_dispatcher = worker_manager

        out_buf = io.StringIO()
        err_buf = io.StringIO()

        def run_graph():
            with redirect_stdout(out_buf), redirect_stderr(err_buf):
                vm.run()

        t = threading.Thread(target=run_graph)
        t.start()
        job = None
        for _ in range(100):
            job = worker_manager.poll(worker_id)
            if job.get("job_id"):
                break
            time.sleep(0.05)
        self.assertIsNotNone(job["job_id"])
        worker_manager.result(worker_id, job["job_id"], "execute")
        t.join(timeout=2)
        self.assertFalse(t.is_alive())

    def test_worker_capability_gpu_only(self):
        from nodus.services.server import WorkerManager
        worker_manager = WorkerManager()
        set_default_dispatcher(worker_manager)
        cpu_worker = worker_manager.register(["cpu"])
        gpu_worker = worker_manager.register(["gpu"])

        src = """
let A = task(fn() { return 1 }, { "worker": "gpu" })
let B = task(fn() { return 2 }, { "worker": "cpu" })
let result = run_graph([A, B])
print(result["tasks"]["task_1"])
print(result["tasks"]["task_2"])
"""
        _loader = ModuleLoader(project_root=None)
        code, functions, code_locs = _loader.compile_only(src, module_name="main.nd")
        vm = lang.VM(code, functions, code_locs=code_locs, source_path="main.nd")
        vm.worker_dispatcher = worker_manager

        def run_graph():
            vm.run()

        t = threading.Thread(target=run_graph)
        t.start()

        cpu_job = self._poll_job(worker_manager, cpu_worker)
        self.assertIsNotNone(cpu_job.get("job_id"))
        self.assertEqual(cpu_job["task_id"], "task_2")
        worker_manager.result(cpu_worker, cpu_job["job_id"], "execute")

        gpu_job = self._poll_job(worker_manager, gpu_worker)
        self.assertIsNotNone(gpu_job.get("job_id"))
        self.assertEqual(gpu_job["task_id"], "task_1")
        worker_manager.result(gpu_worker, gpu_job["job_id"], "execute")

        t.join(timeout=2)
        self.assertFalse(t.is_alive())

    def test_worker_capability_cpu_dispatch(self):
        from nodus.services.server import WorkerManager
        worker_manager = WorkerManager()
        set_default_dispatcher(worker_manager)
        cpu_worker = worker_manager.register(["cpu"])

        src = """
let A = task(fn() { return 3 }, { "worker": "cpu" })
let result = run_graph([A])
print(result["tasks"]["task_1"])
"""
        _loader = ModuleLoader(project_root=None)
        code, functions, code_locs = _loader.compile_only(src, module_name="main.nd")
        vm = lang.VM(code, functions, code_locs=code_locs, source_path="main.nd")
        vm.worker_dispatcher = worker_manager

        t = threading.Thread(target=vm.run)
        t.start()
        job = self._poll_job(worker_manager, cpu_worker)
        self.assertIsNotNone(job.get("job_id"))
        self.assertEqual(job["task_id"], "task_1")
        worker_manager.result(cpu_worker, job["job_id"], "execute")
        t.join(timeout=2)
        self.assertFalse(t.is_alive())

    def test_worker_capability_fallback_dispatch(self):
        from nodus.services.server import WorkerManager
        worker_manager = WorkerManager()
        set_default_dispatcher(worker_manager)
        gpu_worker = worker_manager.register(["gpu"])

        src = """
let A = task(fn() { return 4 }, nil)
let result = run_graph([A])
print(result["tasks"]["task_1"])
"""
        _loader = ModuleLoader(project_root=None)
        code, functions, code_locs = _loader.compile_only(src, module_name="main.nd")
        vm = lang.VM(code, functions, code_locs=code_locs, source_path="main.nd")
        vm.worker_dispatcher = worker_manager

        t = threading.Thread(target=vm.run)
        t.start()
        job = self._poll_job(worker_manager, gpu_worker)
        self.assertIsNotNone(job.get("job_id"))
        self.assertEqual(job["task_id"], "task_1")
        worker_manager.result(gpu_worker, job["job_id"], "execute")
        t.join(timeout=2)
        self.assertFalse(t.is_alive())

    def test_worker_capability_missing(self):
        from nodus.services.server import WorkerManager
        worker_manager = WorkerManager()
        set_default_dispatcher(worker_manager)
        _cpu_worker = worker_manager.register(["cpu"])

        src = """
let A = task(fn() { return 9 }, { "worker": "gpu", "worker_timeout_ms": 50 })
let result = run_graph([A])
print(result["failed"])
print(result["error"])
"""
        _loader = ModuleLoader(project_root=None)
        code, functions, code_locs = _loader.compile_only(src, module_name="main.nd")
        vm = lang.VM(code, functions, code_locs=code_locs, source_path="main.nd")
        vm.worker_dispatcher = worker_manager
        out_buf = io.StringIO()
        err_buf = io.StringIO()
        with redirect_stdout(out_buf), redirect_stderr(err_buf):
            vm.run()
        out = out_buf.getvalue().splitlines()
        self.assertEqual(out[0], "[\"task_1\"]")
        self.assertIn("capability: gpu", out[1])
        events = [e for e in vm.event_bus.events() if e.type == "task_worker_timeout"]
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].name, "task_1")
        self.assertEqual(events[0].data.get("worker"), "gpu")
        self.assertEqual(events[0].data.get("timeout_ms"), 50.0)

    def test_worker_heartbeat_updates_health(self):
        from nodus.services.server import WorkerManager
        worker_manager = WorkerManager()
        worker_manager.event_bus = RuntimeEventBus()
        worker_manager._worker_heartbeat_timeout_ms = 30
        worker_id = worker_manager.register(["cpu"])
        time.sleep(0.02)
        self.assertEqual(worker_manager.heartbeat(worker_id), {"ok": True})
        time.sleep(0.02)
        worker_manager.poll(worker_id)
        self.assertIn(worker_id, worker_manager._workers)
        dead_events = [e for e in worker_manager.event_bus.events() if e.type == "worker_dead"]
        self.assertEqual(dead_events, [])

    def test_worker_death_detection(self):
        from nodus.services.server import WorkerManager
        worker_manager = WorkerManager()
        worker_manager.event_bus = RuntimeEventBus()
        worker_manager._worker_heartbeat_timeout_ms = 10
        worker_id = worker_manager.register(["cpu"])
        time.sleep(0.02)
        worker_manager.poll(worker_id)
        self.assertNotIn(worker_id, worker_manager._workers)
        dead_events = [e for e in worker_manager.event_bus.events() if e.type == "worker_dead"]
        self.assertEqual(len(dead_events), 1)
        self.assertEqual(dead_events[0].data.get("worker_id"), worker_id)

    def test_task_reassignment_after_worker_failure(self):
        from nodus.services.server import WorkerManager
        worker_manager = WorkerManager()
        worker_manager.event_bus = RuntimeEventBus()
        worker_manager._worker_heartbeat_timeout_ms = 20
        set_default_dispatcher(worker_manager)
        worker_a = worker_manager.register(["cpu"])
        worker_b = worker_manager.register(["cpu"])

        src = """
let A = task(fn() { return 10 }, { "worker": "cpu", "worker_timeout_ms": 100 })
let result = run_graph([A])
print(result["tasks"]["task_1"])
"""
        _loader = ModuleLoader(project_root=None)
        code, functions, code_locs = _loader.compile_only(src, module_name="main.nd")
        vm = lang.VM(code, functions, code_locs=code_locs, source_path="main.nd")
        vm.worker_dispatcher = worker_manager

        t = threading.Thread(target=vm.run)
        t.start()

        job_a = self._poll_job(worker_manager, worker_a)
        self.assertIsNotNone(job_a.get("job_id"))
        worker_manager._worker_last_seen[worker_a] = time.monotonic() - 1.0
        worker_manager.poll(worker_b)

        job_b = self._poll_job(worker_manager, worker_b)
        self.assertEqual(job_b.get("job_id"), job_a.get("job_id"))
        worker_manager.result(worker_b, job_b["job_id"], "execute")

        t.join(timeout=2)
        self.assertFalse(t.is_alive())

        requeued = [e for e in worker_manager.event_bus.events() if e.type == "task_requeued"]
        self.assertEqual(len(requeued), 1)
        self.assertEqual(requeued[0].data.get("task_id"), "task_1")

    def test_worker_death_detected_by_sweeper(self):
        from nodus.services.server import RuntimeService
        service = RuntimeService()
        service.workers.event_bus = RuntimeEventBus()
        service.workers._worker_heartbeat_timeout_ms = 10
        worker_id = service.workers.register(["cpu"])
        dead_event = None
        deadline = time.time() + 0.5
        while time.time() < deadline:
            dead_events = [e for e in service.workers.event_bus.events() if e.type == "worker_dead"]
            if dead_events:
                dead_event = dead_events[0]
                break
            time.sleep(0.01)
        self.assertIsNotNone(dead_event)
        self.assertEqual(dead_event.data.get("worker_id"), worker_id)


if __name__ == "__main__":
    unittest.main()



import io
import os
import tempfile
import unittest
from contextlib import redirect_stdout

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader
from cli import debug_file
from nodus.tooling.debugger import Debugger


def make_input(commands: list[str]):
    queued = list(commands)

    def fake_input(_prompt: str) -> str:
        if not queued:
            raise AssertionError("Debugger requested more commands than expected")
        return queued.pop(0)

    return fake_input


class DebuggerTests(unittest.TestCase):
    def test_breakpoint_pauses_on_line(self):
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "main.nd")
            with open(path, "w", encoding="utf-8") as f:
                f.write("let x = 1\nx = x + 1\nprint(x)\n")

            output: list[str] = []
            stdout = io.StringIO()
            with redirect_stdout(stdout):
                exit_code = debug_file(
                    path,
                    debugger_input=make_input([f"break {path}:2", "run", "continue"]),
                    debugger_output=output.append,
                )

            self.assertEqual(exit_code, 0)
            self.assertTrue(any("Breakpoint set at" in line for line in output))
            self.assertTrue(any("paused (breakpoint)" in line for line in output))
            self.assertEqual(stdout.getvalue().splitlines(), ["2.0"])

    def test_stack_inspection_shows_nested_calls(self):
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "stack.nd")
            with open(path, "w", encoding="utf-8") as f:
                f.write(
                    "fn foo() {\n"
                    "    bar()\n"
                    "}\n"
                    "fn bar() {\n"
                    "    let x = 1\n"
                    "    print(x)\n"
                    "}\n"
                    "foo()\n"
                )

            output: list[str] = []
            stdout = io.StringIO()
            with redirect_stdout(stdout):
                exit_code = debug_file(
                    path,
                    debugger_input=make_input([f"break {path}:5", "run", "stack", "continue"]),
                    debugger_output=output.append,
                )

            self.assertEqual(exit_code, 0)
            self.assertTrue(any("main()" in line for line in output))
            self.assertTrue(any("foo()" in line for line in output))
            self.assertTrue(any("bar()" in line for line in output))
            self.assertEqual(stdout.getvalue().splitlines(), ["1.0"])

    def test_variable_inspection_displays_locals(self):
        src = (
            "fn foo(a) {\n"
            "    let x = a + 1\n"
            "    print(x)\n"
            "}\n"
            "foo(4)\n"
        )
        abs_path = os.path.abspath("main.nd")
        _loader = ModuleLoader(project_root=None)
        code, functions, code_locs = _loader.compile_only(src, module_name=abs_path)
        output: list[str] = []
        debugger = Debugger(
            input_fn=make_input([f"break {abs_path}:3", "run", "locals", "print a", "continue"]),
            output_fn=output.append,
            start_paused=True,
        )
        vm = lang.VM(code, functions, code_locs=code_locs, source_path=abs_path, debug=True, debugger=debugger)
        stdout = io.StringIO()
        with redirect_stdout(stdout):
            vm.run()

        self.assertTrue(any('a = 4.0' in line for line in output))
        self.assertTrue(any('x = 5.0' in line for line in output))
        self.assertTrue(any('4.0' == line for line in output))
        self.assertEqual(stdout.getvalue().splitlines(), ["5.0"])

    def test_step_execution_pauses_after_instruction(self):
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "step.nd")
            with open(path, "w", encoding="utf-8") as f:
                f.write("let x = 1\nx = x + 1\nprint(x)\n")

            output: list[str] = []
            stdout = io.StringIO()
            with redirect_stdout(stdout):
                exit_code = debug_file(
                    path,
                    debugger_input=make_input(["step", "continue"]),
                    debugger_output=output.append,
                )

            self.assertEqual(exit_code, 0)
            self.assertTrue(any("paused (step)" in line for line in output))
            self.assertEqual(stdout.getvalue().splitlines(), ["2.0"])


if __name__ == "__main__":
    unittest.main()



import unittest

from nodus.runtime.embedding import NodusRuntime


class EmbeddingApiTests(unittest.TestCase):
    def test_run_source_executes(self):
        rt = NodusRuntime()
        result = rt.run_source('print("hello")', filename="inline.nd")
        self.assertTrue(result["ok"])
        self.assertEqual(result["stdout"].strip(), "hello")

    def test_host_function_registration(self):
        rt = NodusRuntime()

        def add(a, b):
            return a + b

        rt.register_function("add", add)
        result = rt.run_source("print(add(2, 3))", filename="inline.nd")
        self.assertEqual(result["stdout"].strip(), "5")


if __name__ == "__main__":
    unittest.main()
